# import os
# os.chdir("/Users/jameschen")
from PyQt5.QtWidgets import QApplication
# from .lib import *
from .gridGUI import *
from .lib import *
import matplotlib.pyplot as plt
import grid as gd
from grid import lib
import statistics
import cv2
import numpy as np
import pandas as pd
from tqdm import tqdm
app = QApplication(sys.argv)
# === === === === === === DEBUG === === === === === ===

grid = gd.GRID()
grid.loadData("/Users/jameschen/Dropbox/James_Git/FN/data/demo.png")
# grid.loadData("/Users/jameschen/Desktop/testGRID.png")
# grid.loadData(
#     pathImg="/Users/jameschen/Dropbox/James Chen/GRID/Prototype/PineApple.jpg")
grid.binarizeImg(k=3, lsSelect=[0,1], valShad=0, valSmth=0)
# grid.findPlots()
# grid.cpuSeg()

# grid.save(path="/Users/jameschen/", prefix="test")

g = GRID_GUI(grid, 3)  # 0:input, 1:crop, 2:kmean, 3:anchor, 4:output
app.exec_()



# return
# ========== peak searching ==========
# , prominence = (0.01, None)

# def findPeaks(img, nPeaks=0, axis=1, nSmooth=100):
#     """
#     ----------
#     Parameters
#     ----------
#     """
#     # compute 1-D signal
#     signal = img.mean(axis=(not axis)*1) # 0:nrow
#     # ignore signals from iamge frame
#     signal[:2] = 0
#     signal[-2:] = 0
#     # gaussian smooth
#     for _ in range(int(len(signal)/30)):
#         signal = np.convolve(
#             np.array([1, 2, 4, 2, 1])/10, signal, mode='same')
#     # find primary peaks
#     peaks, _ = find_peaks(signal)
#     lsDiff = np.diff(peaks)
#     medSig = statistics.median(signal[peaks])
#     stdSig = np.array(signal[peaks]).std()
#     medDiff = statistics.median(lsDiff)
#     stdDiff = np.array(lsDiff).std()
#     print(lsDiff)
#     print(medDiff)
#     print(stdDiff)
#     # get finalized peaks with distance constrain
#     peaks, _ = find_peaks(signal, distance=medDiff-stdDiff*4)
#     if nPeaks != 0:
#         if len(peaks) > nPeaks:
#             while len(peaks) > nPeaks:
#                 ls_diff = np.diff(peaks)
#                 idx_diff = np.argmin(ls_diff)
#                 idx_kick = idx_diff if (
#                     signal[peaks[idx_diff]] < signal[peaks[idx_diff+1]]) else (idx_diff+1)
#                 peaks = np.delete(peaks, idx_kick)
#         elif len(peaks) < nPeaks:
#             while len(peaks) < nPeaks:
#                 ls_diff = np.diff(peaks)
#                 idx_diff = np.argmax(ls_diff)
#                 peak_insert = (peaks[idx_diff]+peaks[idx_diff+1])/2
#                 peaks = np.sort(np.append(peaks, int(peak_insert)))
#     return peaks, signal


# pks, sig = findPeaks(grid.map.imgs[1])
# plt.plot(sig)
# plt.plot(pks, sig[pks], "x")
# plt.show()

# ========== peak searching ==========


# === === === detect default rank of K === === === ===
# import grid as gd
# import numpy as np
# import matplotlib.pyplot as plt

# grid = gd.GRID()
# grid.loadData("/Users/jameschen/Dropbox/James_Git/FN/data/demo.png")
# # grid.loadData("/Users/jameschen/Dropbox/James Chen/GRID/Modeling/Rhombus.jpg")

# k = 5
# row = k
# col = 4
# i = 0

# grid.binarizeImg(k=k, lsSelect=[0], valShad=0, valSmth=0, outplot=False)

# for i in range(row):
#     imgB = (np.isin(grid.imgs.get('kmean'), i)*1).astype(np.int)
#     sigs = imgB.mean(axis=0)
#     plt.subplot(row, col, 1+i*col+0)
#     plt.imshow(imgB)
#     plt.subplot(row, col, 1+i*col+1)
#     plt.ylim(0, 1)
#     plt.plot(sigs)
#     plt.subplot(row, col, 1+i*col+2)
#     sigf = gd.getFourierTransform(sigs)
#     plt.plot(sigf)
#     plt.subplot(row, col, 1+i*col+3)
#     # scMaxF = round(max(sigf)*5, 4)  # [0, 1]
#     scMaxF = round((max(sigf)/sigf.mean())/100, 4)  # [0, 1]
#     scMean = round(1-(sigs.mean()), 4) # [0, 1]
#     scPeaks = round(len(gd.find_peaks(sigs, height=(sigs.mean()))
#                   [0])/len(gd.find_peaks(sigs)[0]), 4)
#     # scPeaks = len(gd.find_peaks(sigs)[0])
#     score = scMaxF*.25 + scMean*.25 + scPeaks*.5
#     plt.text(.3, .8, str(score), fontsize=10)
#     plt.text(.3, .6, str(scMaxF), fontsize=10)
#     plt.text(.3, .4, str(scMean), fontsize=10)
#     plt.text(.3, .2, str(scPeaks), fontsize=10)

# plt.show()

# === === === detect default rank of K === === === ===


# max(getFourierTransform(sigs))


# plt.imshow()
# plt.show()

# grid = gd.GRID()
# grid.loadData("/Users/jameschen/Dropbox/James Chen/GRID/Modeling/Rhombus.jpg")
# grid.binarizeImg(k=5, lsSelect=[0], valShad=0, valSmth=0, outplot=False)
# grid.findPlots(outplot=False)
# grid.map.dt
# # grid.cpuSeg()
# # grid.save(prefix="Pumptree2")

# app = QApplication(sys.argv)
# g = GRID_GUI(grid, 4) # 0:input, 1:crop, 2:kmean, 3:anchor, 4:output
# app.exec_()


# # ========== kmean ==========
# array = numpy.array([4, 2, 7, 1])
# temp = array.argsort()
# ranks = numpy.empty_like(temp)
# ranks[temp] = numpy.arange(len(array))

# def getRank(array):
#     sort = array.argsort()
#     rank = np.zeros(len(sort), dtype=np.int)
#     rank[sort] = np.flip(np.arange(len(array)), axis=0)
#     return rank

# test = np.array([2, 6, 87,1, 3, 6])
# getRank(test)
# temp = test.argsort()
# np.zeros(len(temp))
# rank = np.empty_like(temp)
# rank[temp] = np.flip(np.arange(len(test)), axis=0)
# np.arange(1, 7, 1)
# np.arange(6, 2, 1)

# ========== plotting ==========

# FIXME: Wrong order for the second axis

# ========== peak searching ==========
# pks, sig = gd.findPeaks(grid.map.imgs[1])
# plt.plot(sig)
# plt.plot(pks, sig[pks], "x")
# plt.show()


# grid.save(prefix="Pumptree2")

# grid = gd.GRID()
# grid.loadData("/Users/jameschen/Dropbox/James Chen/GRID/Modeling/Rhombus.jpg")
# grid.binarizeImg(k=5, lsSelect=[0], valShad=0, valSmth=0, outplot=False)
# grid.findPlots(nRow=7, nCol=6, outplot=False)
# grid.cpuSeg(outplot=True)
# grid.save(prefix="Rhombus")
# # grid.run(pathImg="/Users/jameschen/Dropbox/James Chen/GRID/Modeling/Rhombus.jpg", k=5, lsSelect=[4], path="/Users/jameschen", prefix="letmesleep")

# # with open("/Users/jameschen/Dropbox/photo_grid/Outputter_1018.grid", "wb") as file:
# #     pickle.dump(grid, file, pickle.HIGHEST_PROTOCOL)

# # grid.loadData("/Users/jameschen/Dropbox/James_Git/FN/data/demo.png")
# # grid.binarizeImg(k=3, lsSelect=[0, 1], valShad=0, valSmth=0, outplot=False)


# g = GRID_GUI(grid, 4)
# app.exec_()




# 3rd party imports
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *

# self imports
from ..lib import *
from ..grid import *
from .customQt import * 

class PnCropper(QGroupBox):
    """
    """

    def __init__(self, grid):
        """
        """

        super().__init__()
        self.setStyleSheet("""
        QGroupBox::title{
            subcontrol-origin: margin;
            subcontrol-position: top center;
        }
        """)

        # attr.
        self.grid = grid
        self.layout = QVBoxLayout()
        self.wgImg = Widget_ViewCrop(self.grid.imgs.get('raw'))
        self.initUI()

    def initUI(self):
        """
        """

        self.layout.addWidget(self.wgImg)
        self.setLayout(self.layout)
        self.show()

    def run(self):
        """
        """
        self.grid.cropImg(pts=self.wgImg.getPinnedPoints())

class Widget_ViewCrop(Widget_Img):

    def __init__(self, img):
        super().__init__()
        self.setMouseTracking(True)
        self.img_vis = img
        self.pos = None
        self.zoom = 1
        self.ratio = 1
        self.marks = []
        self.n_marks = 0
        self.imgH, self.imgW = img.shape[0], img.shape[1]
        self.initUI()

    def initUI(self):
        super().make_rgb_img(self.img_vis)
        self.show()

    def paintEvent(self, paint_event):
        painter = QPainter(self)
        super().paintImage(painter)
        pen = QPen()
        pen.setWidth(3)
        pen.setColor(Qt.red)
        painter.setPen(pen)
        painter.setBrush(Qt.white)
        for pos in self.marks:
            drawCross(pos.x(), pos.y(), painter, 5)
        # coordinate
        # if self.pos is not None:
        #     painter.setFont(QFont("Trebuchet MS",14))
        #     painter.drawText(self.pos.x()-20, self.pos.y()+20, "(%d, %d)" %(self.pos.x(), self.pos.y()))
        painter.end()
        
    def mouseMoveEvent(self, event):
        pos = event.pos()
        self.pos = pos
        magArea = int(min(self.rgX[1]-self.rgX[0], self.rgY[1]-self.rgY[0])/5)
        if self.zoom!=0:
            magnifying_glass(self, pos, area=magArea, zoom=self.zoom*1.5)
        else:
            self.setCursor(QCursor(Qt.ArrowCursor))

    def mousePressEvent(self, event):
        if event.button() == Qt.RightButton:
            self.zoom = (self.zoom+1)%3
            self.mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        pt_mouse = event.pos()
        # determine the image boundary
        if event.button() == Qt.LeftButton:
            if self.isFitWidth:
                pt_1d = pt_mouse.y()
                bd_pos = self.rgY[0] + self.sizeImg.height()
                bd_neg = self.rgY[0]
            else:
                pt_1d = pt_mouse.x()
                bd_pos = self.rgX[0] + self.sizeImg.width()
                bd_neg = self.rgX[0]
            # if the click is in the image
            if (pt_1d >= bd_neg) & (pt_1d <= bd_pos):
                if self.n_marks<4:
                    self.marks.append(pt_mouse)
                    self.n_marks += 1
                else:
                    self.marks = []
                    self.n_marks = 0
            # bugmsg('(x:%d, y:%d)' % (pt_mouse.x()*(self.ratio), pt_mouse.y()*(self.ratio)))
            # bugmsg("ratio: %.2f" % (self.ratio))
            self.update()
        self.mouseMoveEvent(event)
        self.pos = None

    def getPinnedPoints(self):
        self.ratio = (self.imgW)/(self.width()) if self.isFitWidth else (self.imgH)/(self.height())
        pts = [[(pt.x()-self.rgX[0])*(self.ratio),
                (pt.y()-self.rgY[0])*(self.ratio)] for pt in self.marks]
        if len(pts)<4:
            pts = [[0, 0], [0, self.imgH], [self.imgW, 0], [self.imgW, self.imgH]]
        return pts



__author__ = "Chunpeng James Chen"
__version__ = "0.3.3"
__update__ = "Feb 20, 2020"

from sys import platform
import subprocess
import json
import sys
from urllib import request
from pkg_resources import parse_version

# welcome message
print("=========*=========*=========*=========*=========*=========")
print("                Welcome to GRID ver.%s!" % __version__)

if "-m" not in sys.argv[0]:
    # if in the command-line environment
    print("    Try 'python -m grid' in Terminel to launch GRID GUI,")
    print("         as command-line version is not ready yet.")

print("=========*=========*=========*=========*=========*=========")

# self update
url = 'https://pypi.python.org/pypi/photo_grid/json'
releases = json.loads(request.urlopen(url).read())['releases']
new_version = sorted(releases, key=parse_version, reverse=True)[0]
if __version__ != new_version:
    # Dialog
    ans = None
    bol_ans = None
    possible_pos_ans = ["y", "Y", "yes"]
    possible_neg_ans = ["n", "N", "no"]

    while bol_ans is None:
        ans = input(
            "A newer version of GRID (ver. %s) is now available, upgrade? (y/n) " % new_version)
        if ans in possible_pos_ans:
            bol_ans = True
        elif ans in possible_neg_ans:
            bol_ans = False

    if bol_ans:
        subprocess.check_call([sys.executable,
                              '-m', 'pip', 'install',
                               'photo_grid', '--upgrade'])
        print("\n\n")
        print("=========*=========*=========*=========*=========*=========")
        print("          Please re-launch GRID to finish the update")
        print("=========*=========*=========*=========*=========*=========")
        quit()


# self imports
print("\n")
from .grid import *

"""
Update Log

- Feb 20, 2020 (0.3.0)
    * Support different display modes in the center detection step
    * Support auto-update feature
    * Bug fixes

- Oct 29, 2019 (0.2.0)
    * Add progress bars
    * Enhance the support for low resolution monitors
    * Improve UI

- Oct 22, 2019 (0.1.3)
    * Optimize default setting of refining parameters
    * Fix wrong angle detection
    * Minor bug fixes

- Oct 19, 2019 (0.1.2)
    * Support rhombus field layout
    * Bug fixes

- Sep 17, 2019 (0.0.16)
    * Improve memory efficiency on Windows OS

- Sep 12, 2019 (0.0.15)
    * Fix problems wiht fixed segmentation
    * Organize file structure
    * Add dark mode
"""



# basic imports
import numpy as np
import pandas as pd
import sys

# self imports
from .io import *
from .lib import *


class GMap():
    """
    """
    def __init__(self):
        """
        ----------
        Parameters
        ----------
        """

        # constant var
        self._degRot = range(-75, 90+1, 15)

        # map file
        self.pdMap = None
        self.isMap = False
        self.countNames = 0

        # dimension
        self.nAxs = [0, 0]
        self.nAxsCur = [0, 0]
        self.nRow = 0
        self.nCol = 0
        self.img = None
        self.imgH = 0
        self.imgW = 0

        # axes
        self.angles = [0, 0]
        self.anglesCur = [0, 0]
        self.slps = [0, 0]  # 0 and 90 degrees for common images
        self.sigs = [0, 0]
        self.itcs = [0, 0]
        self.imgHr = [0, 0]
        self.imgWr = [0, 0]
        self.slpsReset = [0, 0]
        self.itcsReset = [0, 0]

        # iamges for two angle
        self.imgsR_Bin = [0, 0]
        self.imgsR_RGB = [0, 0]

        # output
        self.dt = None

        # progress bar
        self.flag = True
        self.subflag = True
        self.window = 1000

    def load(self, pathMap):
        """
        ----------
        Parameters
        ----------
        """

        self.pdMap = loadMap(pathMap)

        # if self.pdMap is not None:
        #     self.nRow, self.nCol = self.pdMap.shape
        #     self.isMap = True

    def findPlots(self, imgRGB, imgBin, nRow=0, nCol=0, nSmooth=100):
        """
        ----------
        Parameters
        ----------
        """

        # get image info
        self.imgBin = imgBin
        self.imgRGB = imgRGB
        self.imgH, self.imgW = imgBin.shape[:2]
        isDimAssigned = (nRow != 0 and nCol != 0)

        # if dim is assigned regardless have map or not, force changning nR and nC 
        if isDimAssigned:
            self.nAxs = [nCol, nRow]
        elif self.pdMap is not None:
            self.nAxs = [self.pdMap.shape[1], self.pdMap.shape[0]]

        # find angles and slopes
        self.angles = self.detectAngles(img=imgBin, rangeAngle=self._degRot)

        # find intercepts and make centers as pandas DT
        self.locateCenters(nSmooth)

    def detectAngles(self, img, rangeAngle):
        # evaluate each angle
        sc = []
        for angle in rangeAngle:
            imgR = rotateBinNdArray(img, angle)
            sig = imgR.mean(axis=0)
            sigFour = getFourierTransform(sig)
            sc.append(max(sigFour))

        # angle with maximum score win
        scSort = sc.copy()
        scSort.sort()
        idxMax = [i for i in range(len(sc)) if (sc[i] in scSort[-2:])]
        angles = [rangeAngle[idx] for idx in idxMax]

        # return
        return angles

    def locateCenters(self, nSmooth=100):
        self.slps = [1/np.tan(np.pi/180*angle) for angle in self.angles]

        # progress bar
        prog = None
        if self.flag:
            self.flag = False
            prog = initProgress(2, "Calculate slopes and intercepts")

        # find intercepts given 2 angles
        self.updateIntercepts(self.angles, self.nAxs, nSmooth)

        # get pandas data table
        updateProgress(prog, flag=self.subflag)
        self.dt = self.getDfCoordinate(self.angles, self.slps, self.itcs)

        # end progress bar
        if self.subflag and "__main__.py" in sys.argv[0]:
            self.subflag = False
            QTimer.singleShot(
                self.window, lambda: setattr(self, "flag", True))
            QTimer.singleShot(
                self.window, lambda: setattr(self, "subflag", True))

    def updateIntercepts(self, angles, nSigs, nSmooth):
        for i in range(2):
            if (nSigs[i] == 0 or self.nAxsCur[i] != nSigs[i] or self.angles[i] != self.anglesCur[i]):
                imgR_bin, imgR_rgb, sig, intercept = self.cpuIntercept(
                    angles[i], nSigs[i], nSmooth)
                self.imgsR_Bin[i] = imgR_bin
                self.imgsR_RGB[i] = imgR_rgb
                self.imgHr[i] = imgR_bin.shape[0]
                self.imgWr[i] = imgR_bin.shape[1]
                self.sigs[i] = sig
                self.itcs[i] = intercept
                # update number of peaks
                self.nAxs[i] = len(intercept)
                self.nAxsCur[i] = len(intercept)
                # update angles
                self.anglesCur[i] = self.angles[i]

    def cpuIntercept(self, angle, nSig, nSmooth):
        imgR_bin = rotateBinNdArray(self.imgBin, angle)
        imgR_rgb = rotateNdArray(self.imgRGB, angle)
        sig = findPeaks(img=imgR_bin, nPeaks=nSig, nSmooth=nSmooth)[0]
        intercept = getCardIntercept(sig, angle, self.imgH)
        if isinstance(intercept, int):
            intercept = list([intercept])
        return imgR_bin, imgR_rgb, sig, intercept

    def getDfCoordinate(self, angles, slopes, intercepts):
        """
        ----------
        Parameters
        ----------
        """
        imgH, imgW = self.imgBin.shape
        tol = 0.025
        bdN, bdS = -imgH*tol, imgH*(1+tol)
        bdW, bdE = -imgW*tol, imgW*(1+tol)

        idxCol = 0 if abs(slopes[0]) > abs(slopes[1]) else 1
        idxRow = 1 - idxCol
        idxMaj = 0 if getClosedTo0or90(angles[0]) <= getClosedTo0or90(angles[1]) else 1
        idxMin = 1 - idxMaj

        for i in range(2):
            intercepts[i].sort()

        plotsMaj = []
        plotsMin = []
        pts = []

        pMaj = 0
        for itcMaj in intercepts[idxMaj]:  # columns
            pMin = 0
            for itcMin in intercepts[idxMin]:  # rows
                ptX, ptY = solveLines(
                    slopes[idxMaj], itcMaj, slopes[idxMin], itcMin)
                if ptX >= bdW and ptX <= bdE and ptY >= bdN and ptY <= bdS:
                    # outside the frame but within tolerant range
                    if ptX < 0:
                        ptX = 0
                    elif ptX >= imgW:
                        ptX = imgW-1
                    if ptY < 0:
                        ptY = 0
                    elif ptY >= imgH:
                        ptY = imgH-1
                    plotsMaj.append(pMaj)
                    plotsMin.append(pMin)
                    pts.append((ptX, ptY))
                    pMin += 1
            pMaj += 1

        if idxCol==idxMaj:
            dataframe = pd.DataFrame(
                {"row": plotsMin, "col": plotsMaj, "pt": pts})
        elif idxCol==idxMin:
            dataframe = pd.DataFrame(
                {"row": plotsMaj, "col": plotsMin, "pt": pts})

        self.nRow, self.nCol = dataframe['row'].max()+1, dataframe['col'].max()+1

        # update map names
        ctNA = 0
        names = []
        for _, entry in dataframe.iterrows():
            row = entry.row
            col = entry.col
            try:
                names.append(self.pdMap.iloc[row, col])
            except Exception:
                names.append("unnamed_%d" % (ctNA))
                ctNA += 1
        dataframe['var'] = names

        # return
        return dataframe

    def getCoordinate(self, row, col):
        """
        ----------
        Parameters
        ----------
        """
        dt = self.dt
        return dt[(dt.row == row) & (dt.col==col)]['pt'].values[0]

    def getName(self, row, col):
        """
        ----------
        Parameters
        ----------
        """
        dt = self.dt
        return dt[(dt.row == row) & (dt.col == col)]['var'].values[0]

    def delAnchor(self, axis, index):
        # since numpy array required elements in same length if they were
        # temp = self.sigs.tolist()
        temp = self.sigs
        try:
            temp[axis].remove(temp[axis][index])
        except Exception:
            temp[axis] = np.delete(temp[axis], index)
        self.sigs = np.array(temp)
        # temp = self.itcs.tolist()
        temp = self.itcs
        temp[axis] = getCardIntercept(
            self.sigs[axis], self.angles[axis], self.imgH)
        self.itcs = np.array(temp)
        self.dt = self.getDfCoordinate(self.angles, self.slps, self.itcs)

    def addAnchor(self, axis, value):
        # temp = self.sigs.tolist()
        temp = self.sigs
        try:
            temp[axis].append(value)
        except Exception:
            temp[axis] = np.append(temp[axis], value)
        self.sigs = np.array(temp)
        # temp = self.itcs.tolist()
        temp = self.itcs
        temp[axis] = getCardIntercept(
            self.sigs[axis], self.angles[axis], self.imgH)
        self.itcs = np.array(temp)
        self.dt = self.getDfCoordinate(self.angles, self.slps, self.itcs)

    def modAnchor(self, axis, index, value):
        self.sigs[axis][index] = value
        self.itcs[axis] = getCardIntercept(
            self.sigs[axis], self.angles[axis], self.imgH)
        self.dt = self.getDfCoordinate(
            self.angles, self.slps, self.itcs)


def getClosedTo0or90(x):
    s1 = abs(abs(x)-90)
    s2 = abs(abs(x))
    return min(s1, s2)


def scaleTo0and1(array, length):
    array = np.array(array)
    return array/length


def scaleToOrg(array, length):
    array = np.array(array)
    return array*length



# basic imports
import numpy as np

# 3rd party imports
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *

# self imports
from .customQt import *


class PnAnchor(QWidget):
    """
    """

    def __init__(self, grid):
        super().__init__()
        self.grid = grid
        # compute
        self.grid.findPlots()
        self.setFocusPolicy(Qt.StrongFocus)
        self.setFocus()

        # major
        self.layout = QGridLayout()
        self.grRight = QGroupBox()
        self.wgImg = WidgetAnchor(grid)
        self.wgAxs = WidgetAxis()

        self.recImg = QRect(0, 0, 0, 0)
        self.recAxs = QRect(0, 0, 0, 0)
        self.recRight = QRect(0, 0, 0, 0)

        # Right Panel
        self.loRight = QVBoxLayout()
        self.sc_right = QScrollArea()
        self.sc_right.setStyleSheet("QScrollBar {width:0px;}")
        self.sc_right.setWidgetResizable(True)

        # 2 axes
        self.grAxis = [QGroupBox("Axis 1"), QGroupBox("Axis 2")]
        self.loAxis = [QVBoxLayout(), QVBoxLayout()]
        self.grAg = [QGroupBox(), QGroupBox()]
        self.loAg = [QVBoxLayout(), QVBoxLayout()]
        self.dlAg = [QDial(), QDial()]
        self.grNum = [QGroupBox(), QGroupBox()]
        self.loNum = [QHBoxLayout(), QHBoxLayout()]
        self.lbNum = [QLabel("# of peaks"), QLabel("# of peaks")]
        self.spbNum = [QSpinBox(), QSpinBox()]

        self.mtp = 5  # for slider
        self.idxAx = 0

        # Display
        self.gr_dis = QGroupBox("Display")
        self.lo_dis = QHBoxLayout()
        self.rb_bin = QRadioButton("Binary (A)")
        self.rb_rgb = QRadioButton("RGB (S)")

        # reset
        self.btReset = QPushButton("Reset")

        # mouse event
        self.idxAnc = -1
        self.ptX = -1
        self.ptXpress = -1

        # UI
        self.switch = True
        self.initUI()

        # show
        self.show()
        self.grAxis[0].setFocus(True)
        self.wgImg.updateDim()

    def initUI(self):

        # RIGHT: 2 axes
        for i in [0, 1]:
            # config
            angle = self.grid.map.angles[i]
            self.grAg[i].setTitle("Angle: %d degrees" % (angle))
            self.dlAg[i].setRange(-18, 18)
            self.dlAg[i].setValue(int(angle/5))
            self.dlAg[i].setPageStep(3)
            self.dlAg[i].setNotchesVisible(True)
            self.dlAg[i].setNotchTarget(5)
            self.spbNum[i].setValue(len(self.grid.map.itcs[i]))

            # assemble
            self.loAg[i].addWidget(self.dlAg[i])
            self.grAg[i].setLayout(self.loAg[i])

            self.loNum[i].addWidget(self.lbNum[i])
            self.loNum[i].addWidget(self.spbNum[i])
            self.grNum[i].setLayout(self.loNum[i])

            self.loAxis[i].addWidget(self.grAg[i])
            self.loAxis[i].addWidget(self.grNum[i])
            self.grAxis[i].setLayout(self.loAxis[i])
            self.grAxis[i].setCheckable(True)

            self.loRight.addWidget(self.grAxis[i])

        self.grAxis[0].setChecked(True)
        self.grAxis[1].setChecked(False)

        # RIGHT: functions
        self.grAxis[0].clicked.connect(lambda: self.switchAngle(idx=0))
        self.grAxis[1].clicked.connect(lambda: self.switchAngle(idx=1))
        self.dlAg[0].valueChanged.connect(lambda: self.changeAngle(idx=0))
        self.dlAg[1].valueChanged.connect(lambda: self.changeAngle(idx=1))
        self.spbNum[0].valueChanged.connect(lambda: self.changePeaks(idx=0))
        self.spbNum[1].valueChanged.connect(lambda: self.changePeaks(idx=1))
        self.rb_bin.toggled.connect(self.displayImage)
        self.rb_rgb.toggled.connect(self.displayImage)

        # RIGHT: display
        self.rb_bin.setChecked(True)
        self.lo_dis.addWidget(self.rb_bin)
        self.lo_dis.addWidget(self.rb_rgb)
        self.gr_dis.setLayout(self.lo_dis)
        self.loRight.addWidget(self.gr_dis)

        # RIGHT: comp
        self.loRight.addWidget(self.btReset)
        self.grRight.setLayout(self.loRight)
        self.sc_right.setWidget(self.grRight)

        # LEFT IMG: mouse tracking
        # self.wgImg.setMouseTracking(True)

        # Main
        policyRight = QSizePolicy(QSizePolicy.Preferred,
                                  QSizePolicy.Preferred)
        policyRight.setHorizontalStretch(1)
        self.sc_right.setSizePolicy(policyRight)

        policyLeft = QSizePolicy(QSizePolicy.Preferred,
                                 QSizePolicy.Preferred)
        policyLeft.setHorizontalStretch(3)
        self.wgImg.setSizePolicy(policyLeft)
        self.wgAxs.setSizePolicy(policyLeft)

        sizeAxis = 9
        self.layout.addWidget(self.wgImg, 0, 0, sizeAxis, 1)
        self.layout.addWidget(self.wgAxs, sizeAxis, 0, 1, 1)
        self.layout.addWidget(self.sc_right, 0, 1, sizeAxis+1, 1)

        self.setLayout(self.layout)

    def paintEvent(self, event):
        try:
            self.updatePlots()
        except Exception:
            None

        # '''rect'''
        # pen.setWidth(1)
        # pen.setColor(Qt.black)
        # painter.setPen(pen)
        # painter.setBrush(Qt.transparent)
        # painter.drawRect(self.rec_acr_c)
        # painter.drawRect(self.rec_acr_r)

    def updatePlots(self):
        self.updateAnchor()
        self.updateAgents()

    def changeAngle(self, idx):
        # current angle
        angle = self.dlAg[idx].value() * self.mtp
        # oposite angle
        angleOp = self.dlAg[1 - idx].value() * self.mtp
        # angle difference between two axes
        angleDiff = abs(angle - angleOp)
        # if difference is greater than 0
        if angleDiff != 0:
            # if difference is greater than 90 degrees
            if angleDiff > 90:
                if idx == 0:
                    value = min(angle + 90, 90) / self.mtp
                else:
                    value = max(-90, angle - 90) / self.mtp
                self.dlAg[1 - idx].setValue(int(value))
            # if difference is less than 90 degrees and neither of them is 0
            elif angle * angleOp > 0:
                # force the current one equal to 0
                angle = 0
                self.dlAg[idx].setValue(0)
            self.grAg[idx].setTitle("Angle: %d degrees" % (angle))
            self.grid.updateCenters(idx, angle=angle)
            self.switch = False
            self.spbNum[idx].setValue(len(self.grid.map.itcs[idx]))
            self.spbNum[1 - idx].setValue(len(self.grid.map.itcs[1 - idx]))
            self.switch = True
            self.displayImage()
            self.repaint()

    def changePeaks(self, idx):
        if self.switch:
            nPeaks = self.spbNum[idx].value()
            self.grid.updateCenters(idx, nPeaks=nPeaks)
            self.displayImage()
            self.repaint()

    def switchAngle(self, idx):
        self.idxAx = idx
        self.displayImage()
        self.grAxis[idx].setChecked(True)
        self.grAxis[int(not idx)].setChecked(False)
        self.repaint()

    def displayImage(self):
        if self.rb_bin.isChecked():
            self.wgImg.make_bin_img(self.grid.map.imgsR_Bin[self.idxAx])
        else:
            self.wgImg.make_rgb_img(self.grid.map.imgsR_RGB[self.idxAx])
        self.wgImg.repaint()

    def updateAnchor(self):
        ptsRaw = self.grid.map.sigs[self.idxAx]
        rgSrc = (0, self.grid.map.imgWr[self.idxAx])
        rgDst = self.wgImg.rgX
        pts = rescale(ptsRaw, rgSrc, rgDst)
        self.wgAxs.setPoints(pts)

    def updateAgents(self):
        # fetch info
        gmap = self.grid.map
        idxCr = self.idxAx
        idxOp = 1 - idxCr
        agCr = gmap.angles[idxCr]
        agOp = gmap.angles[idxOp]
        agDiff = agOp - agCr
        agAbs = abs(agDiff)

        imgH, imgW = gmap.imgsR_Bin[idxCr].shape
        qimgH, qimgW = self.wgImg.sizeImg.height(), self.wgImg.sizeImg.width()
        ratio = sum([qimgW / imgW, qimgH / imgH]) / 2

        # current axis
        self.wgImg.ptVLine = self.wgAxs.pts

        # another axis
        self.wgImg.slp = 1 / np.tan(np.pi / 180 * agDiff)
        sigs = gmap.sigs[idxOp]

        if agCr % 90 == 0:
            # bugmsg("case 1")
            itc = getCardIntercept(sigs, agDiff, imgH)
        elif agOp % 90 == 0:
            itc = np.cos(np.pi / 180 * agAbs) * gmap.imgHr[idxOp] + \
                    sigs / np.sin(np.pi / 180 * agAbs)
            if agDiff < 0:
                # bugmsg("case 2")
                None
            else:
                # bugmsg("case 3")
                itc = gmap.imgHr[idxCr] - itc
        else:
            seg1 = gmap.imgH * np.cos(np.pi / 180 * abs(agCr))
            seg2 = (sigs - gmap.imgH * np.sin(np.pi / 180 * abs(agOp))) / \
                np.sin(np.pi / 180 * (abs(agOp) + abs(agCr)))
            itc = seg1 + seg2
            if agCr > 0:
                # bugmsg("case 4")
                None
            else:
                # bugmsg("case 5")
                itc = gmap.imgHr[idxCr] - itc

        self.wgImg.itcs = itc * ratio

        # slp = -1/np.tan(np.pi/180*agAbs) if agAbs < 0 else 1/np.tan(np.pi/180*agAbs)
        # sigs = self.grid.map.sigs[idxOp]
        # if idxOp==1:
        #     self.wgImg.slp = slp
        #     X = (sigs/np.sin(np.pi/180*agAbs)) + \
        #         np.cos(np.pi/180*agAbs)*self.grid.map.imgHr[idxOp]
        #     self.wgImg.itcs = (qimgH - X*ratio)
        # else:
        #     self.wgImg.slp = -slp
        #     segA = sigs/np.sin(np.pi/180*agAbs)
        #     segB = np.sin(np.pi/180*agAbs)*self.grid.map.imgWr[idxOp]
        #     self.wgImg.itcs = segA*ratio + (qimgW - segB*ratio)

    def updateDim(self):
        self.recImg = self.wgImg.geometry()
        self.recAxs = self.wgAxs.geometry()
        self.recRight = self.grRight.geometry()

    def getPtGui2Map(self, ptX):
        rgWg = self.wgImg.getImgRange()[0]
        rgMap = (0, self.grid.map.imgWr[self.idxAx]-1)
        return rescale(ptX-self.recAxs.x(), rgWg, rgMap)

    def mousePressEvent(self, event):
        pos = event.pos()
        self.updateDim()
        if self.recImg.contains(pos):
            None  # img
        if self.recAxs.contains(pos):
            self.ptX = self.getPtGui2Map(pos.x())
            self.ptXpress = self.ptX
            self.idxAnc = np.abs(
                self.ptX-self.grid.map.sigs[self.idxAx]).argmin()
        if self.recRight.contains(pos):
            None  # right

    def mouseMoveEvent(self, event):
        pos = event.pos()
        if self.idxAnc != -1:
            if pos.x() > self.wgImg.rgX[0] and pos.x() < self.wgImg.rgX[1]:
                self.ptX = self.getPtGui2Map(pos.x())
            elif pos.x() <= self.wgImg.rgX[0]:
                self.ptX = 0
            elif pos.x() >= self.wgImg.rgX[1]:
                self.ptX = self.grid.map.imgW - 1
            self.grid.map.modAnchor(self.idxAx, self.idxAnc, self.ptX)
            self.update()

    def mouseReleaseEvent(self, event):
        pos = event.pos()
        ptX = self.getPtGui2Map(pos.x())
        sig = np.array(self.grid.map.sigs[self.idxAx])
        if (self.idxAnc != -1 and
           event.button() == Qt.RightButton and
           self.spbNum[self.idxAx].value() > 1):
            self.grid.map.delAnchor(self.idxAx, self.idxAnc)
            value = self.spbNum[self.idxAx].value() - 1
            self.switch = False
            self.spbNum[self.idxAx].setValue(value)
            self.switch = True
        elif (self.ptXpress == ptX and
              event.button() == Qt.LeftButton and
              abs(sig-ptX).min() > sig.std() / 20):
            print("add anchor")
            print("left", abs(sig-ptX).min())
            print("right", sig.std() / 20)
            self.grid.map.addAnchor(self.idxAx, ptX)
            value = self.spbNum[self.idxAx].value() + 1
            self.switch = False
            self.spbNum[self.idxAx].setValue(value)
            self.switch = True

        self.update()
        self.ptX = -1
        self.ptXpress = -1
        self.idxAnc = -1

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_A:
            self.rb_bin.setChecked(True)
        elif event.key() == Qt.Key_S:
            self.rb_rgb.setChecked(True)

    def run(self):
        self.grid.agents.setup(gmap=self.grid.map,
                               img=self.grid.imgs.get('binSeg'))


class WidgetAnchor(Widget_Img):
    def __init__(self, grid):
        super().__init__()
        self.grid = grid
        self.ptVLine = []
        self.itcs = []
        self.slp = 0
        super().make_bin_img(grid.map.imgsR_Bin[0])
        self.repaint()
    # def switch_imgB(self):
    #     super().make_bin_img(self.grid.imgs.get("bin"))
    #     self.repaint()

    # def switch_imgVis(self):
    #     super().make_rgb_img(self.grid.imgs.get("crop"))
    #     self.repaint()

    def paintEvent(self, event):
        painter = QPainter(self)
        super().paintImage(painter)
        pen = QPen()
        pen.setWidth(3)
        pen.setColor(Qt.red)
        painter.setPen(pen)
        painter.setBrush(Qt.white)
        # vertical lines
        for pt in self.ptVLine:
            painter.drawLine(pt, self.rgY[0], pt, self.rgY[1])
            print("paint lines")

        # lines from another axis
        for itc in self.itcs:
            x1, x2 = self.rgX 
            y1 = self.rgY[0] + itc
            y2 = y1 + (x2-x1)*self.slp
            pen.setWidth(1)
            pen.setStyle(Qt.DotLine)
            painter.setPen(pen)
            painter.drawLine(x1, y1, x2, y2)
            pen.setWidth(5)
            painter.setPen(pen)
            for x in self.ptVLine:
                drawCross(x, y1+(x-x1)*self.slp, painter, size=5)
                print("paint lines")

        painter.end()


class WidgetAxis(QWidget):
    def __init__(self):
        super().__init__()
        self.setMouseTracking(True)
        self.pts = []

    def setPoints(self, pts):
        self.pts = pts

    def paintEvent(self, event):
        painter = QPainter(self)
        pen = QPen()
        pen.setWidth(3)
        pen.setColor(Qt.red)
        painter.setPen(pen)
        painter.setBrush(Qt.red)
        # plot triangle
        ptY = int(self.height()/2)
        for ptX in self.pts:
            drawTriangle(ptX, ptY, "North", painter)

        painter.end()


def rescale(values, scaleSrc=(0, 1), scaleDst=(0, 256)):
    values = np.array(values)
    return (values-scaleSrc[0])*(scaleDst[1]-scaleDst[0])/(scaleSrc[1]-scaleSrc[0])+scaleDst[0]



# basic imports
import numpy as np

# self imports
from .io import *
from .lib import *

class GImage():
    """
    """

    def __init__(self):
        """
        ----------
        Parameters
        ----------
        """

        # images
        self.imgs = {
            'raw'    : None,
            'rawRs'  : None,
            'crop'   : None,
            'mean'   : None,
            'kmean'  : None,
            'binOrg' : None,
            'binTemp': None,
            'binSd'  : None,
            'binSm'  : None,
            'bin'    : None,
            'binSeg' : None,
            'visSeg' : None
        }

        # dimension
        self.width, self.height, self.depth = 0, 0, 0
        self.widthRs, self.heightRs = 0, 0
        self.shape = (self.height, self.width, self.depth)
        self.shapeRs = (self.heightRs, self.widthRs, self.depth) 

        # kmean param.
        self.paramKMs = {
            'k' : -1,
            'center': None,
            'rank' : None,
            'features': [],
            'lsSelect' : [],
            'lsKTobin' : [],
            'valShad' : -1,
            'valSmth' : -1
        }

    def get(self, key):
        """
        ----------
        Parameters
        ----------
        """

        return self.imgs[key]

    def set(self, key, value):
        """
        ----------
        Parameters
        ----------
        """

        self.imgs[key] = value

    def load(self, pathImg):
        """
        ----------
        Parameters
        ----------
        """
        if isinstance(pathImg, str):
            # a path to a file
            isLocalImg = pathImg.find("http://") + pathImg.find("https://") == -2

            # image
            if isLocalImg:
                imgInput = loadImg(pathImg)
            else:
                imgInput = loadImgWeb(pathImg)
        elif isinstance(pathImg, np.ndarray):
            # is a numpy
            imgInput = pathImg

        # assign
        self.set(key='raw', value=imgInput)
        self.setShape(shape=imgInput.shape, isRaw=True)

    def getParam(self, key):
        """
        ----------
        Parameters
        ----------
        """

        return self.paramKMs[key]

    def setParam(self, key, value):
        """
        ----------
        Parameters
        ----------
        """

        self.paramKMs[key] = value
    
    def resetParam(self):
        """
        ----------
        Parameters
        ----------
        """

        self.paramKMs = {
            'k': -1,
            'center': None,
            'features': [],
            'lsSelect': [],
            'lsKTobin': [],
            'valShad': -1,
            'valSmth': -1
        }

    def crop(self, pts=None):
        """
        ----------
        Parameters
        ----------
        """

        if pts is None:
            self.set(key='crop',
                     value=self.imgs['raw'])
            self.set(key='mean',
                     value=self.get('crop')[:, :, :3].mean(axis=2))
            self.setShape(shape=self.get(key='crop').shape)
        else:
            self.set(key='crop',
                     value=cropImg(self.imgs['raw'], pts))
            self.set(key='mean',
                     value=self.get('crop')[:, :, :3].mean(axis=2))
            self.setShape(shape=self.get(key='crop').shape)

        self.resetParam()

    def doKMeans(self, k=3, features=[0, 1, 2], colorOnly=False):
        """
        ----------
        Parameters
        ----------
        """

        if len(features) == 0:
            print("no feature is selected")
        elif (k != self.paramKMs['k']) or\
             (features != self.paramKMs['features']):
            # Will skip if no updates on the params
            imgK, center = doKMeans(img=self.get('crop'),
                                    k=k,
                                    features=features)
            self.set(key='kmean', value=imgK)
            # update parameters
            self.paramKMs['center'] = center
            self.paramKMs['rank'] = self.rankCenters(
                k=k, center=center, colorOnly=colorOnly)
            # can't update yet as binarize needs it
            # self.paramKMs['k'] = k 
            # self.paramKMs['features'] = features
        else:
            # skip
            bugmsg("skip kmean")

    def binarize(self, k=3, features=[0, 1, 2], lsSelect=[0]):
        """
        ----------
        Parameters
        ----------
        """

        if len(features) == 0:
            print("no feature is selected")
        elif (k != self.paramKMs['k']) or\
             (features != self.paramKMs['features']) or\
             (lsSelect != self.paramKMs['lsSelect']):
            # ratioK = [(self.paramKMs['center'][i, 0]-self.paramKMs['center'][i, 1])/self.paramKMs['center'][i, :].sum()
            #           for i in range(self.paramKMs['center'].shape[0])]
            # # rankK = np.flip(np.argsort(ratioK), 0)
            # rankK = np.argsort(ratioK)
            try:
                clusterSelected = self.paramKMs['rank'][lsSelect]
            except Exception:
                clusterSelected = []
            self.set(key='binOrg', value=(
                (np.isin(self.get('kmean'), clusterSelected))*1).astype(np.int))
            self.set(key='binTemp', value=self.get('binOrg').copy())
            self.set(key='binSm', value=self.get('binOrg').copy())
            # udpate parameters
            self.paramKMs['k'] = k
            self.paramKMs['features'] = features
            self.paramKMs['lsSelect'] = lsSelect
            self.paramKMs['lsKToBin'] = clusterSelected
        else:
            # skip
            bugmsg("skip binarize")

    def smooth(self, value):
        """
        ----------
        Parameters
        ----------
        """

        if value != self.paramKMs['valSmth']:
            valSmthDiff = value - self.paramKMs['valSmth']
            if valSmthDiff > 0:
                valSmthReal = valSmthDiff
            else:
                valSmthReal = value
                self.set(key='binTemp', value=self.get(key='binOrg').copy())      
            self.set(key='binTemp', value=smoothImg(image=self.get(key='binTemp'), n=valSmthReal))
            self.set(key='binSm',   value=binarizeSmImg(self.get(key='binTemp')))
            # update parameters
            self.paramKMs['valSmth'] = value
        else:
            # skip
            bugmsg("skip smoothing")

    def deShadow(self, value):
        """
        ----------
        Parameters
        ----------
        """

        if value != self.paramKMs['valShad']:
            self.set(key='binSd', value=(self.get(key='mean')>=value)*1)
            # update parameter
            self.paramKMs['valShad'] = value
        else:
            # skip
            bugmsg("skip shadow")

    def finalized(self):
        """
        ----------
        Parameters
        ----------
        """

        self.set(key='bin', value=np.multiply(
            self.get('binSm'), self.get('binSd')))

    def readyForSeg(self):
        """
        ----------
        Parameters
        ----------
        """
        nSmt = int(min(self.get('bin').shape[0], self.get('bin').shape[1])/300)
        self.set(key='binSeg', value=blurImg(self.get('bin'),
                                             n=nSmt))
        # compute the vis/seg image
        imgTemp = np.expand_dims(self.get("bin"), axis=2)
        imgSeg = np.multiply(self.get('crop')[:, :, :3], imgTemp).copy()
        imgSeg[(imgSeg.mean(axis=2) == 0), :] = 255
        self.set(key='visSeg', value=imgSeg)

    def rankCenters(self, k, center, colorOnly=False):
        scores = []
       
        if colorOnly:
            ratioK = [(center[i, 0]-center[i, 1])/center[i, :].sum()
                    for i in range(center.shape[0])]
            rank = np.flip(np.argsort(ratioK), axis=0)
        else:
            for i in range(k):
                imgB = (np.isin(self.get('kmean'), i)*1).astype(np.int)
                sigs = imgB.mean(axis=0)
                sigsF = getFourierTransform(sigs)
                scMaxF = round((max(sigsF)/sigsF.mean())/100, 4)  # [0, 1]
                scMean = round(1-(sigs.mean()), 4)  # [0, 1]
                try:
                    scPeaks = round(len(find_peaks(sigs, height=(sigs.mean()))
                                        [0])/len(find_peaks(sigs)[0]), 4)
                except Exception:
                    scPeaks = 0

                score = scMaxF*.25 + scMean*.25 + scPeaks*.5
                scores.append(score)
            rank = np.flip(np.array(scores).argsort(), axis=0)

        return rank

    def rotate(self, nRot):
        """
        ----------
        Parameters
        ----------
        """

        for key in self.imgs.keys():
            bugmsg(key)
            if key == 'raw' or key == 'rawRs': continue
            try:
                self.set(key=key, value=np.rot90(self.get(key=key), nRot))
            except Exception as e:
                bugmsg(e)
                None
        self.setShape(self.get("crop").shape)

    def setShape(self, shape, isRaw=False):
        """
        ----------
        Parameters
        ----------
        """
        if isRaw:
            try:
                self.height, self.width, self.depth = shape
            except:
                self.height, self.width = shape
                self.depth = 1
            self.shape = (self.height, self.width, self.depth)
        else:
            try:
                self.heightRs, self.widthRs, self.depth = shape
            except:
                self.heightRs, self.widthRs = shape
                self.depth = 1
            self.shapeRs = (self.heightRs, self.widthRs, self.depth)

    def getShape(self, is3D=False, isRaw=False):
        """
        ----------
        Parameters
        ----------
        """
        if is3D:
            return self.shape if isRaw else self.shapeRs
        else:
            return self.shape[:2] if isRaw else self.shapeRs[:2] 



import platform
import os


class GUser():
    """
    """
    def __init__(self):
        """
        """
        self.platform = platform.system()
        self.architecture = platform.architecture()[0]
        self.release = platform.release()
        self.machine = platform.machine()
        self.dirHome = os.path.expanduser("~")
        self.dirGrid = os.path.split(__file__)[0]

    def printInfo(self):
        print("GRID User's Info")
        print("----------------")
        print("Platform:      ",  self.platform)
        print("Architecture:  ",  self.architecture)
        print("Release:       ",  self.release)
        print("Machine:       ",  self.machine)
        print("Home Dir:      ",  self.dirHome)
        print("GRID Dir:      ",  self.dirGrid)



# basic imports 
from enum import Enum

# 3rd party imports
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *

# self imports
from .grid import *
from .gui import * 
from .lib import *

class GRID_GUI(QMainWindow):
    """
    """

    def __init__(self, gridInput=None, idxPn=None):
        """
        ----------
        Parameters
        ----------
        """
        super().__init__()
        #         QWidget {
        #     font: 16pt Trebuchet MS
        # }
        self.setStyleSheet("""
        QWidget {
            font: 16pt Trebuchet MS
        }
        QGroupBox::title{
            subcontrol-origin: margin;
            left: 10px;
            padding: 0 3px 0 3px;
        }
        QGroupBox {
            border: 1px solid gray;
            border-radius: 9px;
            margin-top: 0.5em;
        }
        """)

        # CLI
        self.grid = GRID()

        # GUI
        self.nPanel = -1
        self.pnContent = QWidget()
        self.pnMain = QStackedWidget()

        self.pnNavi = QWidget()
        self.btNext = QPushButton()
        self.btPrev = QPushButton()
        self.prog = GProg(size=5, name="Specify an image to proceed", widget=self.pnContent)
        self.layout = None

        # Shortcut
        self.gridInput = gridInput
        self.idxPn = idxPn

        # UI
        self.initUI()    

    def startover(self):
        while self.pnMain.count()!=0:
            widget = self.pnMain.widget(self.pnMain.count()-1)
            self.pnMain.removeWidget(widget)

        self.nPanel = -1
        self.gridInput = None
        self.idxPn = None
        del self.grid
        self.grid = GRID()
        self.initUI()

    def initUI(self):
        """
        ----------
        Parameters
        ----------
        """

        # window setup
        self.setWindowTitle("GRID")
        self.resize(1280, 700)
        self.centerWindow()

        if self.gridInput is None:
            # initialize with first panel
            self.showInputer()
        else:
            # or use shortcut
            self.grid = self.gridInput
            if self.idxPn==0:
                self.showInputer()
            elif self.idxPn==1:
                self.showCropper()
            elif self.idxPn==2:
                self.showKMeaner()
            elif self.idxPn==3:
                self.showAnchor()
            elif self.idxPn==4:
                self.showOutputer()

        # show
        self.show()

    def showInputer(self, isNew=True):
        bugmsg("show input")
        self.prog.set(n=0, name="Specify an image to proceed")
        self.assembleNavigation(nameNext="Load Files ->", oneSide=True)
        self.btNext.clicked.connect(
            lambda: self.showCropper())
        self.updateMainPn(panel=Panels.INPUTER, isNew=isNew)

    def showCropper(self, isNew=True):
        bugmsg("crop")
        self.prog.set(n=1, name="Click on the image to specify FOUR cornors of the area of interest (AOI)")
        self.assembleNavigation()
        self.btPrev.clicked.connect(
            lambda: self.showInputer(isNew=False))
        self.btNext.clicked.connect(
            lambda: self.showKMeaner())
        self.updateMainPn(panel=Panels.CROPPER, isNew=isNew)     

    def showKMeaner(self, isNew=True):
        bugmsg("kmean")
        self.prog.set(
            n=2, name="Define the pixels of interest (POI)")
        self.assembleNavigation()
        self.btPrev.clicked.connect(
            lambda: self.showCropper(isNew=False))
        self.btNext.clicked.connect(
            lambda: self.showAnchor())
        self.updateMainPn(panel=Panels.KMEANER, isNew=isNew)

    def showAnchor(self, isNew=True):
        bugmsg("anchor")
        self.prog.set(
            n=3, name="Define the plot centers: LEFT-CLICK to add, RIGHT-CLICK to remove, DRAG to adjust")
        self.assembleNavigation()
        self.btPrev.clicked.connect(
            lambda: self.showKMeaner(isNew=False))
        self.btNext.clicked.connect(
            lambda: self.showOutputer())
        self.updateMainPn(panel=Panels.ANCHOR, isNew=isNew)

    def showOutputer(self, isNew=True):
        bugmsg("show output")
        self.prog.set(
            n=4, name="Finalize the segmentation and export results: Drag to adjust borders")
        self.assembleNavigation(nameNext="Finish")
        self.btPrev.clicked.connect(
            lambda: self.showAnchor(isNew=False))
        self.btNext.clicked.connect(lambda: self.finalize())
        self.updateMainPn(panel=Panels.OUTPUTER, isNew=isNew)

    def updateMainPn(self, panel, isNew=True):
        # traverse forward
        if isNew:
            try:
                # run computation from the previous panel
                bugmsg("run")
                self.pnMain.currentWidget().run()
            except Exception as e:
                print(e)
                # except the initial one
                None

            self.pnMain.addWidget(panel.value[1](self.grid))
            self.nPanel += 1
        # traverse backward
        else:
            widget = self.pnMain.widget(panel.value[0]+1)
            self.pnMain.removeWidget(widget)
            self.nPanel -= 1

        # set current widget
        self.pnMain.setCurrentIndex(self.nPanel)

        # show
        self.assembleAndShow()

    def finalize(self):
        """
        ----------
        Parameters
        ----------
        """

        msgBox = QMessageBox()
        msgBox.setIcon(QMessageBox.Information)
        msgBox.setText("Finished!")
        msgBox.setInformativeText("Save and start another job?")
        msgBox.setStandardButtons(QMessageBox.Save | QMessageBox.Discard | QMessageBox.Cancel)
        returnValue = msgBox.exec()
        if returnValue == QMessageBox.Save:
            path = self.pnMain.currentWidget().fd_output.text()
            prefix = self.pnMain.currentWidget().fd_project.text()
            isH5 = self.pnMain.currentWidget().ck_h5.isChecked()
            self.grid.save(path=path, prefix=prefix, h5=isH5)
            self.startover()
        elif returnValue == QMessageBox.Discard:
            self.startover()

    def centerWindow(self):
        """
        ----------
        Parameters
        ----------
        """

        center = QApplication.desktop().availableGeometry().center()
        rect = self.geometry()
        rect.moveCenter(center)
        self.setGeometry(rect)

    def assembleNavigation(self, nameNext="Next ->", namePrev="<- Prev", oneSide=False):
        """
        ----------
        Parameters
        ----------
        """

        self.pnNavi = QWidget()
        self.btNext = QPushButton(nameNext)
        self.btPrev = QPushButton(namePrev)
        loNavi = QHBoxLayout()
        if oneSide:
            loNavi.addStretch(1)
        else:
            loNavi.addWidget(self.btPrev)
        loNavi.addWidget(self.btNext)
        self.pnNavi.setLayout(loNavi)

    def assembleAndShow(self):
        """
        ----------
        Parameters
        ----------
        """

        self.layout = QVBoxLayout()
        self.layout.setSpacing(0)
        self.layout.addWidget(self.prog)
        self.layout.addWidget(self.pnMain, Qt.AlignCenter)
        self.layout.addWidget(self.pnNavi)
        self.pnContent = QWidget()
        self.pnContent.setLayout(self.layout)
        self.setCentralWidget(self.pnContent)
        self.setMinimumHeight(30)
        self.show()


class Panels(Enum):
    INPUTER = 0, PnInputer
    CROPPER = 1, PnCropper
    KMEANER = 2, PnKmeaner
    ANCHOR = 3, PnAnchor
    OUTPUTER = 4, PnOutputer



# ARCHIVE:
# def test(self):
#     import json
#     with open('anchors', 'w') as fout:
#         json.dump(self.params['anchors'], fout)
#     np.save("img_crop", self.params['crop'])
#     np.save("img_bin", self.params['bin'])
#     np.save("map", self.params['map'])
#     np.save("img_k", self.params['k'])
#     np.save("ls_bin", self.params['ls_bin'])
#     bugmsg("nc:%d" % (self.params['nc']))
#     bugmsg("nr:%d" % (self.params['nr']))


# def updateMainPn(self, panel, isNew=True):
#     # traverse forward
#     if isNew:
#         try:
#             # run computation from the previous panel
#             bugmsg("run")
#             self.pnMain.currentWidget().run()
#         except:
#             # except the initial one
#             None
#         self.pnMain.addWidget(panel.value[1](self.grid))
#     # traverse backward
#     else:
#         self.pnMain.removeWidget(self.pnMain.widget(panel.value[0]+1))

#     # set current widget
#     self.pnMain.setCurrentIndex(panel.value[0])

#     # show
#     self.assembleAndShow()
# 
# def updateMainPn(self, panel, isNew=True):
#     """
#     ----------
#     Parameters
#     ----------
#     """
#     # define events
#     if panel == Panels.INPUTER:
#         bugmsg("panel Input")
#         self.assembleNavigation(nameNext="Load Files ->", oneSide=True)
#         self.btNext.clicked.connect(
#             lambda: self.updateMainPn(Panels.CROPPER))
#     elif panel == Panels.CROPPER:
#         bugmsg("panel cropper")
#         self.assembleNavigation()
#         self.btPrev.clicked.connect(
#             lambda: self.updateMainPn(Panels.INPUTER, isNew=False))
#         self.btNext.clicked.connect(
#             lambda: self.updateMainPn(Panels.KMEANER))
#     elif panel == Panels.KMEANER:
#         bugmsg("panel kmeaner")
#         self.assembleNavigation()
#         self.btPrev.clicked.connect(
#             lambda: self.updateMainPn(Panels.CROPPER, isNew=False))
#         self.btNext.clicked.connect(
#             lambda: self.updateMainPn(Panels.ANCHOR))
#     elif panel == Panels.ANCHOR:
#         self.assembleNavigation()
#         self.btPrev.clicked.connect(
#             lambda: self.updateMainPn(Panels.KMEANER, isNew=False))
#         self.btNext.clicked.connect(
#             lambda: self.updateMainPn(Panels.OUTPUTER))
#     elif panel == Panels.OUTPUTER:
#         self.assembleNavigation(nameNext="Finish")
#         self.btPrev.clicked.connect(
#             lambda: self.updateMainPn(Panels.ANCHOR, isNew=False))
#         self.btNext.clicked.connect(lambda: self.finalize())

#     # traverse forward
#     if isNew:
#         try:
#             # run computation from the previous panel
#             bugmsg("run")
#             self.pnMain.currentWidget().run()
#         except:
#             # except the initial one
#             None
#         self.pnMain.addWidget(panel.value[1](self.grid))
#     # traverse backward
#     else:
#         self.pnMain.removeWidget(self.pnMain.widget(panel.value[0]+1))

#     # set current widget
#     self.pnMain.setCurrentIndex(panel.value[0])

#     # show
#     self.assembleAndShow()
# 



# basic imports
import os
import sys
import warnings
warnings.filterwarnings("ignore")

# self imports
from .guser import *
from .gimage import *
from .gmap import *
from .gagent import *
from .lib import *


class GRID():

    def __init__(self):
        """
        ----------
        Parameters
        ----------
        """

        # self-defined class
        self.user = GUser()
        self.imgs = GImage()
        self.map = GMap()
        self.agents = GAgent()

        # for progressbar
        self.flag = True
        self.subflag = True
        self.window = 1000

    def __user__(self):
        """
        ----------
        Parameters
        ----------
        """

        self.user.printInfo()

    def run(self, pathImg=None, pathMap=None, pts=None,
            k=3, features=[0, 1, 2], lsSelect=[0], valShad=0, valSmth=0,
            nRow=0, nCol=0, nSmooth=100,
            tol=5, coefGrid=.2,
            outplot=False,
            path=None, prefix="GRID",
            preset=None):
        """
        ----------
        Parameters
        ----------
        """

        if preset is not None:
            params = getPickledGRID(preset)
            self.run(pathImg=pathImg, pathMap=pathMap, pts=pts,
                     nSmooth=nSmooth,
                     tol=tol,
                     **params, outplot=outplot)
        else:
            prog = initProgress(6, "loading data")
            self.loadData(pathImg=pathImg, pathMap=pathMap, outplot=outplot)
            prog = updateProgress(prog, 1, "cropping")
            self.cropImg(pts=pts, outplot=outplot)
            prog = updateProgress(prog, 1, "binarizing")
            self.binarizeImg(k=k, features=features,
                             lsSelect=lsSelect, 
                             valShad=valShad, valSmth=valSmth, outplot=outplot)
            prog = updateProgress(prog, 1, "locating plots")
            self.findPlots(nRow=nRow, nCol=nCol,
                           nSmooth=nSmooth, outplot=outplot)
            prog = updateProgress(prog, 1, "segmenting")
            self.cpuSeg(tol=tol, coefGrid=coefGrid, outplot=outplot)
            prog = updateProgress(prog, 1, "exporting")
            self.save(path=path, prefix=prefix)
            prog = updateProgress(prog, 1, "done!")

    def save(self, path=None, prefix="GRID", h5=False):
        """
        ----------
        Parameters
        ----------
        """
        if "__main__.py" not in sys.argv[0]:
            app = QApplication(sys.argv)

        self.savePlotAndDT(path=path, prefix=prefix, h5=h5)

        if "__main__.py" not in sys.argv[0]:
            app.quit()

        params = {
            "k": self.imgs.paramKMs["k"],
            "features": self.imgs.paramKMs["features"],
            "lsSelect": self.imgs.paramKMs["lsSelect"],
            "valShad":  self.imgs.paramKMs["valShad"],
            "valSmth":  self.imgs.paramKMs["valSmth"],
            "nRow": self.agents.nRow,
            "nCol": self.agents.nCol,
            "coefGrid": self.agents.coef
        }

        try:
            pathOut = os.path.join(path, prefix) + ".grid"
            pickleGRID(params, pathOut)
        except Exception:
            pathOut = os.path.join(self.user.dirHome, prefix) + ".grid"
            pickleGRID(params, pathOut)

    # === === === === === === MAJOR WORKFLOW === === === === === ===

    def loadData(self, pathImg=None, pathMap=None, outplot=False):
        """
        ----------
        Parameters
        ----------
        """
        if pathImg is None:
            self.imgs.load(
                pathImg=os.path.join(self.user.dirGrid, "demo/seg_img.jpg"))
            self.map.load(
                pathMap=os.path.join(self.user.dirGrid, "demo/seg_map.csv"))
        else:
            self.imgs.load(pathImg=pathImg)
            self.map.load(pathMap=pathMap)

        if outplot:
            pltImShow(self.imgs.get("raw")[:, :, :3])

    def cropImg(self, pts=None, outplot=False):
        """
        ----------
        Parameters
        ----------
        """
        print(pts)
        self.imgs.crop(pts)

        if outplot:
            pltImShow(self.imgs.get("crop")[:, :, :3])

    def binarizeImg(self, k=3, features=[0, 1, 2], lsSelect=[0], valShad=0,
                    valSmth=0, colorOnly=False, outplot=False):
        """
        ----------
        Parameters
        ----------
        """
        # from QtCore import QTimer

        if self.imgs.get("crop") is None:
            self.cropImg()

        # KMEAN
        prog = None
        if self.flag:
            self.flag = False
            prog = initProgress(5, name="K-Means Clustering")
        self.imgs.doKMeans(k=k, features=features, colorOnly=colorOnly)

        # BINARIZE
        updateProgress(prog, name="Binarizing", flag=self.subflag)
        self.imgs.binarize(k=k, features=features, lsSelect=lsSelect)

        # SMOOTH
        updateProgress(prog, name="Smoothing", flag=self.subflag)
        self.imgs.smooth(value=valSmth)

        # DESHADOW
        updateProgress(prog, name="DeShade-ing", flag=self.subflag)
        self.imgs.deShadow(value=valShad)

        # FINALIZE
        updateProgress(prog, name="Finalizing", flag=self.subflag)
        self.imgs.finalized()
        updateProgress(prog, name="Done", flag=self.subflag)

        # set progress bar inactive for 300ms
        if self.subflag and "__main__.py" in sys.argv[0]:
            self.subflag = False
            QTimer.singleShot(self.window, lambda: setattr(self, "flag", True))
            QTimer.singleShot(self.window, lambda: setattr(self, "subflag", True))

        # Plot
        if outplot:
            pltImShowMulti(
                imgs=[self.imgs.get('crop')[:, :, :3],
                      self.imgs.get('kmean'),
                      self.imgs.get('binOrg'),
                      self.imgs.get('bin')],
                titles=["Original", "K-Means", "Binarized", "Finalized"])

    def findPlots(self, nRow=0, nCol=0, nSmooth=100, outplot=False):
        """
        ----------
        Parameters
        ----------
        """

        # iamge
        self.imgs.readyForSeg()

        self.map.findPlots(imgRGB=self.imgs.get("crop")[:, :, :3],
                           imgBin=self.imgs.get("binSeg"),
                           nRow=nRow, nCol=nCol, nSmooth=nSmooth)

        self.agents.setup(gmap=self.map, img=self.imgs.get('binSeg'))

        if outplot:
            pltLinesPlot(gmap=self.map, agents=self.agents.agents,
                         img=self.imgs.get('binSeg'))

    def cpuSeg(self, tol=5, coefGrid=.2, outplot=False):
        """
        ----------
        Parameters
        ----------
        """

        self.agents.cpuPreDim(tol=tol)
        self.agents.autoSeg(coefGrid=coefGrid)

        if outplot:
            pltSegPlot(self.agents, self.imgs.get("visSeg"), isRect=True)

    # === === === === === === IMAGE === === === === === ===

    def rotateImg(self, nRot):
        """
        ----------
        Parameters
        ----------
        """
        self.imgs.rotate(nRot)

    # === === === === === === MAP === === === === === ===

    def updateCenters(self, idx, angle=-1, nPeaks=0):
        if angle != -1:
            self.map.angles[idx] = angle
        if nPeaks != 0:
            self.map.nAxs[idx] = nPeaks
        self.map.locateCenters()
        # self.map.nAxs[idx] = 0

    # === === === === === === AGENTS === === === === === ===

    def fixSeg(self, width, length):
        """
        ----------
        Parameters
        ----------
        """
        self.agents.fixSeg(width, length)

    # === === === === === === OUTPUT === === === === === ===

    def savePlotAndDT(self, path=None, prefix="GRID", h5=False):
        """
        ----------
        Parameters
        ----------
        """
        if path is None or not os.path.exists(path):
            path = self.user.dirHome

        proh5 = 1 if h5 else 0
        # progress bar
        prog = initProgress(2 + proh5, "Exporting Dataframe")

        saveDT(self, path, prefix)
        updateProgress(prog, name="Exporting Figures")
        savePlot(self, path, prefix)

        if h5:
            updateProgress(prog, name="Exporting h5 file")
            saveH5(self, path, prefix)

        updateProgress(prog, name="Done")



# basic imports
import sys
import os

# 3rd party imports
from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import QTimer
import qdarkstyle

# self imports
from .gridGUI import *

if "--test" not in sys.argv:
    app = QApplication(sys.argv)
    if '--light' not in sys.argv:
        app.setStyleSheet(qdarkstyle.load_stylesheet_pyqt5())

    grid = GRID_GUI()

    timer = QTimer()
    timer.timeout.connect(lambda: None)
    timer.start(100)

    app.exec_()

else:
    # TEST MODE
    from .gtest import *



import setuptools
setuptools.setup(name='photo_grid',
      version='0.3.3',
      description='A GUI for field segmentation',
      url='https://github.com/Poissonfish/photo_grid',
      python_requires='>=3.6',
      classifiers=[
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows :: Windows 10"
      ],
      author='James Chen',
      author_email='chun-peng.chen@wsu.edu',
      license='GPLv3',
      packages=['grid', 'grid.gui'],
      include_package_data=True,
      install_requires=['numpy', 'pandas>=0.19.2',
                        # data processing
                        'h5py',
                        # math, models
                        'sklearn', 'scipy', 'matplotlib',
                        # image processing
                        'image', 'opencv-python==4.1.1.26', 'rasterio',
                        # GUI
                        'PyQt5==5.12', 'qdarkstyle',
                        # misc
                        'tqdm'])



# Basic imports
import io
import os
import numpy as np
import pandas as pd
from urllib.request import urlopen
from PIL import Image

# 3-rd party imports
from PyQt5.QtCore import QFile, QIODevice
import h5py
import rasterio

# self import
from .lib import initProgress, updateProgress, pltImShow, pltSegPlot
from .dir import Dir


def loadImg(path):
    """
    ----------
    Parameters
    ----------
    path : str
           path to the image file

    -------
    Returns
    -------
    npImg : 3-d ndarray encoded in UINT8

    """

    # rasObj = rasterio.open(path)
    # nCh = rasObj.count
    # obj = initProgress(nCh, "loading channel 1")
    # if nCh < 3:
    #     npImg = np.zeros((rasObj.height, rasObj.width, 3), dtype="uint8")
    #     for i in range(3):
    #         npImg[:, :, i] = rasObj.read(1)
    #         updateProgress(obj, 1, "loading channel %d" %
    #                        (i+2 if i != (nCh-1) else i+1))
    # else:
    #     npImg = np.zeros((rasObj.height, rasObj.width, nCh), dtype="uint8")
    #     for i in range(nCh):
    #         npImg[:, :, i] = rasObj.read(i + 1)
    #         updateProgress(obj, "loading channel %d" % (i+2) if i != (nCh-1)
    # e
    # lse "Done")

    # return npImg
    rasObj = rasterio.open(path)
    nCh = rasObj.count
    prog = initProgress(nCh, name="Loading channel 1")
    if nCh < 3:
        npImg = np.zeros((rasObj.height, rasObj.width, 3), dtype="uint8")
        for i in range(3):
            npImg[:, :, i] = rasObj.read(1)
            if i != nCh-1:
                updateProgress(prog, name="Loading channel %d" % (i+2))
            else:
                updateProgress(prog, name="Done")
    else:
        npImg = np.zeros((rasObj.height, rasObj.width, nCh), dtype="uint8")
        for i in range(nCh):
            npImg[:, :, i] = rasObj.read(i + 1)
            if i != nCh-1:
                updateProgress(prog, name="Loading channel %d" % (i+2))
            else:
                updateProgress(prog, name="Done")

    return npImg


def loadImgWeb(URL):
    """
    ----------
    Parameters
    ----------
    URL : str
          URL to the UINT8-encoded image file

    -------
    Returns
    -------
    npImg : 3-d ndarray encoded in UINT8

    """

    with urlopen(URL)as url:
        file = io.BytesIO(url.read())
        npImg = np.array(Image.open(file), dtype="uint8")

    return npImg


def loadMap(path):
    """
    ----------
    Parameters
    ----------
    path : str
           path to the csv file

    -------
    Returns
    -------
    pdMap : Pandas dataframe or None if path is empty

    """

    try:
        pdMap = pd.read_csv(path, header=None)
    except Exception:
        pdMap = None

    return pdMap


def saveQImg(qimg, path):
    """
    ----------
    Parameters
    ----------
    qimg : qimage

    path : str
           path to the destination
    -------
    Returns
    -------
    None

    """

    qfile = QFile(path + ".jpg")
    qfile.open(QIODevice.WriteOnly)
    qimg.save(qfile, "JPG")
    qfile.close()


def saveDT(grid, path, prefix="GRID"):
    # get path
    pathDT = os.path.join(path, prefix+"_data.csv")

    # progress bar
    nD = grid.imgs.depth
    lsK = grid.imgs.paramKMs["lsSelect"]

    # grab info from GRID obj
    img = grid.imgs.get("crop").copy().astype(np.int)
    ch1Sub = 1 if img.shape[2] == 3 else 3  # replace NIR with Gr if it's RGB

    # intialize dataframe
    df = pd.DataFrame(columns=['var', 'row', 'col',
                               'area_all', 'area_veg'])

    # calculate index imgs
    dicIdx = dict({
        "NDVI": (img[:, :, ch1Sub] - img[:, :, 0]) /
                (img[:, :, ch1Sub] + img[:, :, 0] + 1e-8),
        "GNDVI": (img[:, :, ch1Sub] - img[:, :, 1]) /
                (img[:, :, ch1Sub] + img[:, :, 1] + 1e-8),
        "CNDVI": (2 * img[:, :, ch1Sub] - img[:, :, 0] - img[:, :, 1]) /
                (img[:, :, ch1Sub] + img[:, :, 0] + img[:, :, 1] + 1e-8),
        "RVI": img[:, :, ch1Sub] / (img[:, :, 0] + 1e-8),
        "GRVI": img[:, :, ch1Sub] / (img[:, :, 1] + 1e-8),
        "NDGI": (img[:, :, 1] - img[:, :, 0]) /
                (img[:, :, 1] + img[:, :, 0] + 1e-8)
    })

    # channel values
    for i in range(nD):
        name = "ch_%d" % i
        dicIdx[name] = img[:, :, i]

    # ratio of each cluster (k)
    cluster = 0
    for k in lsK:
        name = "cluster_%d" % cluster
        dicIdx[name] = (np.isin(grid.imgs.get("kmean"), i))*1
        cluster += 1
        
    # append columns based on the dict
    for key, _ in dicIdx.items():
        df[key] = None
        df[key + "_std"] = None

    # index data frame
    for row in range(grid.agents.nRow):
        for col in range(grid.agents.nCol):
            agent = grid.agents.get(row, col)
            if not agent or agent.isFake():
                continue
            entry = dict(var=agent.name, row=row, col=col)

            # get ROI region
            rg_row = range(agent.getBorder(Dir.NORTH),
                           agent.getBorder(Dir.SOUTH))
            rg_col = range(agent.getBorder(Dir.WEST),
                           agent.getBorder(Dir.EAST))

            # get selected pixels info
            imgBinAgent = grid.imgs.get('bin')[rg_row, :][:, rg_col]
            n_veg = imgBinAgent.sum()

            # load area info
            entry["area_all"] = len(rg_row)*len(rg_col)
            entry["area_veg"] = n_veg

            # append temp entry
            for key, imgIdx in dicIdx.items():
                imgIdxAgent = imgIdx[rg_row, :][:, rg_col]
                img_out = np.multiply(imgBinAgent, imgIdxAgent)
                vec_out = img_out[img_out != 0].flatten()
                entry[key] = vec_out.mean()
                entry[key + "_std"] = vec_out.std()

            df.loc[len(df)] = entry

    df = df[~df['var'].isnull()]

    # export
    df.to_csv(pathDT, index=False)


def savePlot(grid, path, prefix="GRID"):
    pltImShow(grid.imgs.get("crop")[:, :, :3],
              path=path, prefix=prefix, filename="_raw.png")

    pltSegPlot(grid.agents, grid.imgs.get("crop")[:, :, :3],
               isRect=True,
               path=path, prefix=prefix, filename="_rgb.png")

    pltImShow(grid.imgs.get("kmean"),
              path=path, prefix=prefix, filename="_kmeans.png")

    pltSegPlot(grid.agents, grid.imgs.get("visSeg"),
               isRect=True,
               path=path, prefix=prefix, filename="_seg.png")

    pltSegPlot(grid.agents, grid.imgs.get("bin"),
               isRect=True,
               path=path, prefix=prefix, filename="_bin.png")


def saveH5(grid, path, prefix="GRID"):
    # get path
    pathH5 = os.path.join(path, prefix+".h5")

    # create file
    with h5py.File(pathH5, "w"):
        None  # create file

    # index data frame
    img = grid.imgs.get("crop").copy()
    for row in range(grid.agents.nRow):
        for col in range(grid.agents.nCol):
            agent = grid.agents.get(row, col)
            if not agent or agent.isFake():
                continue
            key = agent.name

            # get ROI region
            rgY = range(agent.getBorder(Dir.NORTH),
                        agent.getBorder(Dir.SOUTH))
            rgX = range(agent.getBorder(Dir.WEST),
                        agent.getBorder(Dir.EAST))

            # compute kernel
            imgAll = img[:, rgX, :][rgY, :, :]
            imgBin = grid.imgs.get("bin")[:, rgX][rgY, :]
            imgFin = np.multiply(imgAll, np.expand_dims(imgBin, 2))

            # export image
            try:
                with h5py.File(pathH5, "a") as f:
                    f.create_dataset(key, data=imgFin, compression="gzip")
            except Exception:
                print("Failed to save %s" % key)




# basic imports
import numpy as np
import sys
import time
import math
import os
import pickle
import statistics

# 3rd party imports
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from tqdm import tqdm, tqdm_gui
import cv2
from scipy.signal import convolve2d
from scipy.signal import find_peaks
import matplotlib.pyplot as plt
from matplotlib import patches
from matplotlib.lines import Line2D


def doKMeans(img, k=3, features=[0]):
    """
    ----------
    Parameters
    ----------
    """

    # data type conversion for opencv
    ## select features
    img = img[:, :, features].copy()
    ## standardize
    img_max, img_min = img.max(axis=(0, 1)), img.min(axis=(0, 1))-(1e-8)
    img = (img-img_min)/(img_max-img_min)
    ## convert to float32
    img_z = img.reshape((-1, img.shape[2])).astype(np.float32)
    
    # define criteria, number of clusters(K) and apply kmeans()
    criteria = (cv2.TERM_CRITERIA_MAX_ITER, 10, 1.0)
    param_k = dict(data=img_z,
                   K=k,
                   bestLabels=None,
                   criteria=criteria,
                   attempts=10,
                   flags=cv2.KMEANS_PP_CENTERS)

    # KMEANS_RANDOM_CENTERS
    _, img_k_temp, center = cv2.kmeans(**param_k)

    # Convert back
    img_k = img_k_temp.astype(np.uint8).reshape((img.shape[0], -1))

    # return
    return img_k, center


def blurImg(image, n, cutoff=0.5):
    image = smoothImg(image=image, n=n)
    return binarizeSmImg(image, cutoff=cutoff)


def smoothImg(image, n):
    """
    ----------
    Parameters
    ----------
    """

    kernel = np.array((
            [1, 4, 1],
            [4, 9, 4],
            [1, 4, 1]), dtype='int')/29

    for _ in range(n):
        image = convolve2d(image, kernel, mode='same')

    return image


def binarizeSmImg(image, cutoff=0.5):
    """
    ----------
    Parameters
    ----------
    """
    imgOut = image.copy()
    imgOut[image > cutoff] = 1
    imgOut[image <= cutoff] = 0

    return imgOut.astype(np.int)


def cropImg(img, pts):
    """
    ----------
    Parameters
    ----------
    img : str
          path to the image file 
    pts : list of 2-tuple
          each tuple is a coordinate (x, y)
    -------
    Returns
    -------
    npImg : 3-d ndarray encoded in UINT8

    """

    # convert to opencv cpmatitable
    pts = np.array(pts).astype(np.float32)
    # find four corners
    token = True
    while token:
        try:
            order_x = np.argsort([pts[i, 0] for i in range(4)])
            order_y = np.argsort([pts[i, 1] for i in range(4)])
            pt_NW = pts[order_x[:2][np.isin(order_x[:2], order_y[:2])][0]]
            pt_SW = pts[order_x[:2][np.isin(order_x[:2], order_y[2:])][0]]
            pt_NE = pts[order_x[2:][np.isin(order_x[2:], order_y[:2])][0]]
            pt_SE = pts[order_x[2:][np.isin(order_x[2:], order_y[2:])][0]]
            token = False
        except:
            pts = rotatePts(pts, 15)
    # generate sorted source point
    pts = np.array([pt_NW, pt_NE, pt_SW, pt_SE])
    # estimate output dimension
    img_W = (sum((pt_NE-pt_NW)**2)**(1/2)+sum((pt_SE-pt_SW)**2)**(1/2))/2
    img_H = (sum((pt_SE-pt_NE)**2)**(1/2)+sum((pt_SW-pt_NW)**2)**(1/2))/2
    while (img_W > 1600):
        img_W /= 2
        img_H /= 2
    shape = (int(img_W), int(img_H))
    # generate target point
    pts2 = np.float32(
        [[0, 0], [shape[0], 0], [0, shape[1]], [shape[0], shape[1]]])
    # transformation
    M = cv2.getPerspectiveTransform(pts, pts2)
    dst = cv2.warpPerspective(img, M, (shape[0], shape[1]))
    dst = np.array(dst).astype(np.uint8)

    return dst


def rotatePts(pts, angle):
    """
    ----------
    Parameters
    ----------
    """
    ptx = np.array([pts[i, 0] for i in range(len(pts))])
    pty = np.array([pts[i, 1] for i in range(len(pts))])
    qx = math.cos(math.radians(angle))*(ptx) - \
        math.sin(math.radians(angle))*(pty)
    qy = math.sin(math.radians(angle))*(ptx) + \
        math.cos(math.radians(angle))*(pty)
    qpts = [[qx[i], qy[i]] for i in range(len(pts))]
    return np.array(qpts)

# === === === === === GRID pickle === === === === ===


def pickleGRID(obj, path):
    with open(path, "wb") as file:
        pickle.dump(obj, file, pickle.HIGHEST_PROTOCOL)


def getPickledGRID(path):
    with open(path, "rb") as file:
        obj = pickle.load(file)

    return obj

# === === === === === rank === === === === ===


def getRank(array):
    """
    ----------
    Results
    ----------
    [1,3,6,2,4] -> [4,2,0,3,1]
    """
    sort = np.array(array).argsort()
    rank = np.zeros(len(sort), dtype=np.int)
    rank[sort] = np.flip(np.arange(len(array)), axis=0)
    return rank

# === === === === === peak searching === === === === ===
def rotateNdArray(img, angle):
    depth = img.shape[2]
    list_img = []

    for i in range(depth):
        imgTemp = img[:, :, i]
        if imgTemp.max() > 1:
            val_max = 255
        else:
            val_max = 1

        # create border for the image
        imgTemp[:, 0:2] = val_max
        imgTemp[0:2, :] = val_max
        imgTemp[:, -2:] = val_max
        imgTemp[-2:, :] = val_max

        # padding
        sizePad = max(imgTemp.shape)
        imgP = np.pad(imgTemp, [sizePad, sizePad], 'constant')

        # rotate
        pivot = tuple((np.array(imgP.shape[:2])/2).astype(np.int))
        matRot = cv2.getRotationMatrix2D(pivot, -angle, 1.0)
        imgR = cv2.warpAffine(
            imgP.astype(np.float32), matRot, imgP.shape, flags=cv2.INTER_LINEAR).astype(np.uint8)

        # crop
        sigX = np.where(imgR.sum(axis=0) != 0)[0]
        sigY = np.where(imgR.sum(axis=1) != 0)[0]
        imgC = imgR[sigY[0]:sigY[-1], sigX[0]:sigX[-1]]

        # store output img
        list_img.append(imgC)

    imgFinal = np.array(list_img)
    imgFinal = np.swapaxes(imgFinal, 0, 1)
    imgFinal = np.swapaxes(imgFinal, 1, 2)

    return imgFinal

def rotateBinNdArray(img, angle):
    # create border for the image
    img[:, 0:2] = 1
    img[0:2, :] = 1
    img[:, -2:] = 1
    img[-2:, :] = 1

    # padding
    sizePad = max(img.shape)
    imgP = np.pad(img, [sizePad, sizePad], 'constant')

    # rotate
    pivot = tuple((np.array(imgP.shape[:2])/2).astype(np.int))
    matRot = cv2.getRotationMatrix2D(pivot, -angle, 1.0)
    imgR = cv2.warpAffine(
        imgP.astype(np.float32), matRot, imgP.shape, flags=cv2.INTER_LINEAR).astype(np.uint8)

    # crop
    sigX = np.where(imgR.sum(axis=0) != 0)[0]
    sigY = np.where(imgR.sum(axis=1) != 0)[0]
    imgC = imgR[sigY[0]:sigY[-1], sigX[0]:sigX[-1]]

    # return
    return imgC


def rotateVec(vec, angle):
    deg = np.pi/180
    x, y = vec[0], vec[1]
    xp = np.cos(deg*angle)*x - np.sin(deg*angle)*y
    yp = np.sin(deg*angle)*x + np.cos(deg*angle)*y
    return (xp, yp)


def getFourierTransform(sig):
    sigf = abs(np.fft.fft(sig)/len(sig))
    return sigf[2:int(len(sigf)/2)]
    # return sigf[2:25]


def getCardIntercept(lsValues, angle, imgH=0):
    if angle == 0:
        return lsValues * 1
    else:
        coef = 1 / np.sin(np.pi / 180 * abs(angle))
        if angle < 0:
            return lsValues*coef
        else:
            return imgH-lsValues*coef


def getLineABC(slope, intercept):
    if np.isinf(slope):
        A = 1
        B = 0
        C = intercept
    else:
        A = slope
        B = -1
        C = -intercept
    return A, B, C


def solveLines(slope1, intercept1, slope2, intercept2):
    A1, B1, C1 = getLineABC(slope1, intercept1)
    A2, B2, C2 = getLineABC(slope2, intercept2)
    D = A1*B2-A2*B1
    Dx = C1*B2-B1*C2
    Dy = A1*C2-C1*A2
    if D != 0:
        x, y = Dx/D, Dy/D
        return int(x), int(y)
    else:
        return False, False


def findPeaks(img, nPeaks=0, axis=1, nSmooth=100):
    """
    ----------
    Parameters
    ----------
    """

    # compute 1-D signal
    signal = img.mean(axis=(not axis)*1) # 0:nrow

    # ignore signals from iamge frame
    signal[:2] = [0, 0]
    signal[-2:] = [0, 0]

    # gaussian smooth 
    for _ in range(int(len(signal)/30)):
        signal = np.convolve(
            np.array([1, 2, 4, 2, 1])/10, signal, mode='same')

    # find primary peaks
    peaks, _ = find_peaks(signal)
    lsDiff = np.diff(peaks)
    medSig = statistics.median(signal[peaks])
    stdSig = np.array(signal[peaks]).std()
    medDiff = statistics.median(lsDiff)
    stdDiff = np.array(lsDiff).std()

    # get finalized peaks with distance constrain
    coef = 0.18/(stdDiff/medDiff) # empirical 
    peaks, _ = find_peaks(signal, distance=medDiff-stdDiff*coef)
    # , prominence=(0.01, None))
    if nPeaks != 0:
        if len(peaks) > nPeaks:
            while len(peaks) > nPeaks:
                ls_diff = np.diff(peaks)
                idx_diff = np.argmin(ls_diff)
                idx_kick = idx_diff if (
                    signal[peaks[idx_diff]] < signal[peaks[idx_diff+1]]) else (idx_diff+1)
                peaks = np.delete(peaks, idx_kick)
        elif len(peaks) < nPeaks:
            while len(peaks) < nPeaks:
                ls_diff = np.diff(peaks)
                idx_diff = np.argmax(ls_diff)
                peak_insert = (peaks[idx_diff]+peaks[idx_diff+1])/2
                peaks = np.sort(np.append(peaks, int(peak_insert)))

    return peaks, signal


# === === === === Plotting === === === === ===


def pltCross(x, y, size=3, width=1, color="red"):
    pt1X = [x-size, x+size]
    pt1Y = [y-size, y+size]
    line1 = Line2D(pt1X, pt1Y, linewidth=width, color=color)
    pt2X = [x-size, x+size]
    pt2Y = [y+size, y-size]
    line2 = Line2D(pt2X, pt2Y, linewidth=width, color=color)    
    return line1, line2


def pltImShow(img, path=None, prefix="GRID", filename=".png"):
    if path is None:
        ax = plt.subplot(111)
        ax.imshow(img)
        plt.show()
    else:
        file = os.path.join(path, prefix+filename)
        if img.max()==1:
            qimg = getBinQImg(img)
        elif img.max() < 100:
            qimg = getIdx8QImg(img, img.max()+1)
        else:
            qimg = getRGBQImg(img)

        qimg.save(file, "PNG")


def pltSegPlot(agents, plotBase, isRect=False, path=None, prefix="GRID", filename=".png"):
    if path is None:
        ax = plt.subplot(111)
        ax.imshow(plotBase)
        for row in range(agents.nRow):
            for col in range(agents.nCol):
                agent = agents.get(row=row, col=col)
                if not agent:
                    continue
                recAg = agent.getQRect()
                line1, line2 = pltCross(agent.x, agent.y, width=1)
                ax.add_line(line1)
                ax.add_line(line2)
                if isRect:
                    rect = patches.Rectangle(
                        (recAg.x(), recAg.y()), recAg.width(), recAg.height(),
                        linewidth=1, edgecolor='r', facecolor='none')
                    ax.add_patch(rect)
        plt.show()
    else:
        file = os.path.join(path, prefix+filename)
        qimg = getBinQImg(plotBase) if plotBase.max() == 1 else getRGBQImg(plotBase)

        pen = QPen()
        pen.setWidth(3)
        pen.setColor(Qt.red)
        painter = QPainter(qimg)
        painter.setPen(pen)
        painter.setBrush(Qt.transparent)
        for row in range(agents.nRow):
            for col in range(agents.nCol):
                agent = agents.get(row, col)
                rect = agent.getQRect()
                painter.drawRect(rect)
        painter.end()
        qimg.save(file, "PNG")


def pltImShowMulti(imgs, titles=None, vertical=False):
    nImgs = len(imgs)
    idxImg = 100 if vertical else 10
    idxLyt = 20 if vertical else 200

    plt.figure()
    for i in range(nImgs):
        idxPlot = idxImg*round(nImgs/2) + idxLyt   (i+1)
        plt.subplot(idxPlot)
        plt.imshow(imgs[i])
        try:
            plt.title(titles[i])
        except Exception:
            None

    plt.show()


def pltLinesPlot(gmap, agents, img):
    itcs = gmap.itcs
    slps = gmap.slps
    # plotting
    _, ax = plt.subplots()
    ax.imshow(img)
    for i in range(2):
        for intercept in itcs[i]:
            plotLine(ax, slps[i], intercept)
    for cAgents in agents:
        for agent in cAgents:
            line1, line2 = pltCross(agent.x, agent.y, width=2)
            ax.add_line(line1)
            ax.add_line(line2)

    plt.show()


def plotLine(axes, slope, intercept):
    if abs(slope) > 1e+9:
        # vertical line
        y_vals = np.array(axes.get_ylim())
        x_vals = np.repeat(intercept, len(y_vals))
    else:
        # usual line
        x_vals = np.array(axes.get_xlim())
        y_vals = intercept + slope * x_vals
    axes.plot(x_vals, y_vals, '--', color="red")


def bugmsg(msg, title="DEBUG"):
    if "--test" in sys.argv:
        print("======%s=====" % title)
        print(msg)


# for GUI
def getRGBQImg(img):
    h, w = img.shape[0], img.shape[1]
    qImg = QImage(img.astype(np.uint8).copy(), w, h, w*3, QImage.Format_RGB888)
    return QPixmap(qImg)


def getBinQImg(img):
     h, w = img.shape[0], img.shape[1]
     qImg = QImage(img.astype(np.uint8).copy(), w, h, w*1, QImage.Format_Indexed8)
     qImg.setColor(0, qRgb(0, 0, 0))
     qImg.setColor(1, qRgb(241, 225, 29))
     return QPixmap(qImg)


def getIdx8QImg(img, k):
    colormap = [qRgb(228, 26, 28),
                qRgb(55, 126, 184),
                qRgb(77, 175, 74),
                qRgb(152, 78, 163),
                qRgb(255, 127, 0),
                qRgb(255, 255, 51),
                qRgb(166, 86, 40),
                qRgb(247, 129, 191),
                qRgb(153, 153, 153)]
    h, w = img.shape[0], img.shape[1]
    qImg = QImage(img.astype(np.uint8).copy(), w, h, w*1, QImage.Format_Indexed8)
    for i in range(k):
        qImg.setColor(i, colormap[i])
    return QPixmap(qImg)


def getGrayQImg(img):
    h, w = img.shape[0], img.shape[1]
    qImg = QImage(img.astype(np.uint8).copy(), w, h, w*1, QImage.Format_Grayscale8)
    return QPixmap(qImg)


# progress bar
class GProg(QWidget):
    def __init__(self, size, name, widget):
        super().__init__()
        try:
            self._width = widget.width()/5
            self._height = self._width/16*5
            wgW, wgH = widget.width(), widget.height()
            self._pos = widget.pos()
        except Exception:
            self._width, self._height = 1, 1
            wgW, wgH = 1, 1
            self._pos = QPoint(1, 1)

        self.label = QLabel(name)
        font = QFont("Trebuchet MS", 20)
        self.label.setFont(font)
        self.bar = QProgressBar()
        self.bar.setRange(0, size)
        self.bar.setValue(0)
        self.layout = QVBoxLayout()
        self.layout.addWidget(self.label)
        self.layout.addWidget(self.bar)
        self.setLayout(self.layout)
        self.move(self._pos.x()+(wgW-self._width)/2,
                  self._pos.y()+(wgH-self._height)/2)
        self.resize(self._width, self._height)
        self.show()
        self.repaint()
        QApplication.processEvents()

    def inc(self, n, name=None):
        self.bar.setValue(self.bar.value()+n)
        if name is not None:
            self.label.setText(name)
        self.repaint()
        QApplication.processEvents()

    def set(self, n, name=None):
        self.bar.setValue(n)
        if name is not None:
            self.label.setText(name)
        self.repaint()
        QApplication.processEvents()


def initProgress(size, name=None):
    if "__main__.py" in sys.argv[0]:
        # GUI
        widget = QApplication.activeWindow()
        obj = GProg(size, name, widget)
    else:
        # CLT
        obj = tqdm(total=size, postfix=name)

    return obj


def updateProgress(obj, n=1, name=None, flag=True):
    if (not flag) or (obj is None):
        return 0

    if "__main__.py" in sys.argv[0]:
        # GUI
        obj.inc(n, name)
    else:
        # CLT
        obj.set_postfix_str(name)
        obj.update(n)



# basic imports
import numpy as np

# self imports
from .lib import *
from .dir import Dir


class GAgent():

    def __init__(self):
        """
        ----------
        Parameters
        ----------
        """

        self.agents = []
        self.img = None
        self.shape = (0, 0)
        self.imgH, self.imgW = self.shape
        self.nRow, self.nCol = 0, 0
        self.rowFake, self.colFake = -1, -1
        self.coef = 0

        # progress bar
        self.flag = True
        self.subflag = True
        self.window = 1000

    def setup(self, gmap, img):
        """
        ----------
        Parameters
        ----------
        """

        self.agents = []
        self.img = img
        self.shape = self.img.shape
        self.imgH, self.imgW = self.shape
        self.nRow, self.nCol = gmap.nRow, gmap.nCol
        dt = gmap.dt
        fr, fc = [], []

        for row in range(self.nRow):
            lsAgentsRow = []
            for col in range(self.nCol):
                try:
                    entry = dt[(dt.row == row) & (dt.col == col)].iloc[0]                
                    ptX, ptY = entry["pt"]
                    name = entry["var"]
                except Exception:
                    # handle inconsistant # of rows/cols
                    fr.append(row)
                    fc.append(col)
                    ptX, ptY = (-1, -1)
                    name = "_FAKE"
                agent = Agent(name=name, row=row, col=col)
                self.setCoordinate(agent=agent, x=ptX, y=ptY)
                lsAgentsRow.append(agent)
            self.agents.append(lsAgentsRow)

        self.rowFake, self.colFake = -1, -1
        # determine fake plots
        if len(fr) > 1 and np.array(fr).var() == 0:
            self.rowFake = fr[0]
        elif len(fc) > 1 and np.array(fc).var() == 0:
            self.colFake = fc[0]

    def get(self, row, col):
        """
        ----------
        Parameters
        ----------
        """

        if row < 0 or col < 0:
            # negative index
            return False
        else:
            try:
                return self.agents[row][col]
            except Exception:
                # outside frame
                return False

    def getNeib(self, row, col, dir):
        """
        ----------
        Parameters
        ----------
        """

        if dir == Dir.EAST:
            agent = self.get(row, col+1)
            if not agent:
                # outside image
                return False
            elif (col+1 == self.colFake):
                # on fake line
                return agent
            elif agent.isFake():
                return self.get(row-1, col+1)
            else:
                return agent
        elif dir == Dir.SOUTH:
            agent = self.get(row+1, col)
            if not agent:
                # outside image
                return False
            elif (row+1 == self.rowFake):
                # on fake line
                return agent
            elif agent.isFake():
                return self.get(row+1, col-1)
            else:
                return agent
        elif dir == Dir.WEST:
            agent = self.get(row, col-1)
            if not agent:
                # outside image
                return False
            elif agent.isFake():
                return self.get(row-1, col-1)
            else:
                return agent
        elif dir == Dir.NORTH:
            agent = self.get(row-1, col)
            if not agent:
                # outside image
                return False
            elif agent.isFake():
                return self.get(row+1, col-1)
            else:
                return agent

    def cpuPreDim(self, tol=5):
        """
        ----------
        Parameters
        ----------
        """

        # prog = initProgress(self.nRow, name="Estimating dimentions")
        for row in range(self.nRow):
            # updateProgress(prog)
            for col in range(self.nCol):
                agentSelf = self.get(row, col)
                agentSelf.resetBorderHard()
                if not agentSelf or agentSelf.isFake():
                    continue
                rgTemp = dict()
                for axis in [0, 1]:
                    # extract direction info and 1dImg
                    dir1 = Dir(axis) # axis:0, return N(0) and S(2)
                    dir2 = Dir(axis+2) # axis:1, return W(1) and E(3)
                    axisRev = (not axis)*1
                    img1d = self.img[agentSelf.y,
                                     :] if axis else self.img[:, agentSelf.x]
                    # extract agents info
                    ptSelf = agentSelf.getCoordinate()[axisRev]
                    agentNeig1 = self.getNeib(row, col, dir1)
                    agentNeig2 = self.getNeib(row, col, dir2)
                    # if both neighbors exists
                    if (agentNeig1 != 0) and (agentNeig2 != 0) and (not agentNeig1.isFake()) and (not agentNeig2.isFake()):
                        ptNeig1 = agentNeig1.getCoordinate()[axisRev]
                        ptNeig2 = agentNeig2.getCoordinate()[axisRev]
                        ptMid = int((ptNeig1+ptNeig2)/2)
                        ptBd1 = int((ptNeig1+ptMid)/2)
                        ptBd2 = int((ptNeig2+ptMid)/2)
                    # if only left/up side exist
                    elif agentNeig1:
                        ptNeig1 = agentNeig1.getCoordinate()[axisRev]
                        ptBd1 = int((ptSelf+ptNeig1)/2)
                        ptBd2 = self.img.shape[axis]
                        self.setBorder(agentSelf, dir2, ptBd2)
                    # if only right/down side exist
                    elif agentNeig2:
                        ptNeig2 = agentNeig2.getCoordinate()[axisRev]
                        ptBd1 = 0
                        ptBd2 = int((ptSelf+ptNeig2)/2)
                        self.setBorder(agentSelf, dir1, ptBd1)
                    # if neither side exist
                    elif self.nRow==1 or self.nCol==1:
                        ptBd1 = 0
                        ptBd2 = self.img.shape[axis]
                        self.setBorder(agentSelf, dir1, ptBd1)
                        self.setBorder(agentSelf, dir2, ptBd2)

                    # get predim
                    ## negative side (neighber 1)
                    pt_cur = ptSelf
                    tol_cur = 0
                    while (tol_cur < tol) & (pt_cur > ptBd1):
                        try:
                            img_val = img1d[pt_cur]
                        except:
                            break
                        tol_cur += 1 if img_val == 0 else -tol_cur #else reset to 0
                        pt_cur -= 1
                    rgTemp[dir1.name] = pt_cur
                    ## positive side (neighber 2)
                    pt_cur = ptSelf
                    tol_cur = 0
                    while (tol_cur < tol) & (ptBd2 > pt_cur):
                        try:
                            img_val = img1d[pt_cur]
                        except:
                            break
                        tol_cur += 1 if img_val == 0 else -tol_cur #else reset to 0
                        pt_cur += 1
                    rgTemp[dir2.name] = pt_cur
                agentSelf.setPreDim(rgTemp)

    def autoSeg(self, coefGrid=.2):
        """
        ----------
        Parameters
        ----------
        """
        self.coef = coefGrid
        
        # reset the border first
        self.resetBorder()

        # progress bar
        prog = None
        if self.flag:
            self.flag = False
            prog = initProgress(self.nRow, name="Segmenting")

        # loop over rows and cols
        for row in range(self.nRow):
            updateProgress(prog, flag=self.subflag)
            for col in range(self.nCol):
                agentSelf = self.get(row, col)
                for dir in list([Dir.EAST, Dir.SOUTH]):
                    agentNeib = self.getNeib(row, col, dir)
                    dirNeib = list(Dir)[(dir.value+2) %
                                         4]  # reverse the direction
                    if not agentSelf or agentSelf.isFake() or not agentNeib or agentNeib.isFake():
                        continue
                    # reset agent border
                    agentSelf.border[dir.name] = agentSelf.border_reset[dir.name]
                    agentNeib.border[dirNeib.name] = agentNeib.border_reset[dirNeib.name]
                    # calculate border
                    dist_agents = abs(
                        agentSelf.x-agentNeib.x) if dir == Dir.EAST else abs(agentSelf.y-agentNeib.y)
                    while abs(agentSelf.getBorder(dir)-agentNeib.getBorder(dirNeib)) > 1:
                        scASelf = agentSelf.getScoreArea(
                            dir, self.img)
                        scGSelf = agentSelf.getScoreGrid(
                            dir)/dist_agents
                        scANeib = agentNeib.getScoreArea(
                            dirNeib, self.img)
                        scGNeib = agentNeib.getScoreGrid(
                            dirNeib)/dist_agents
                        # except Exception as e:
                        scoreSelf = scASelf - (scGSelf*coefGrid)
                        scoreNeib = scANeib - (scGNeib*coefGrid)
                        if scoreSelf > scoreNeib:                            
                            self.updateBorder(agentSelf, dir, 1)
                        elif scoreSelf < scoreNeib:
                            self.updateBorder(agentNeib, dirNeib, -1)
                        else:
                            self.updateBorder(agentSelf, dir, 1)
                            self.updateBorder(agentNeib, dirNeib, -1)

        if self.subflag and "__main__.py" in sys.argv[0]:
            self.subflag = False
            QTimer.singleShot(self.window, lambda: setattr(self, "flag", True))
            QTimer.singleShot(self.window, lambda: setattr(self, "subflag", True))

        # rescue plots on fake line
        if self.rowFake != -1:
            for col in range(1, self.nCol):
                agent = self.get(self.rowFake, col)
                if not agent.isFake():
                    agentNeig = self.get(self.rowFake-1, col-1)
                    agent.setBorder(Dir.WEST, agentNeig.getBorder(Dir.EAST))
        elif self.colFake != -1:
            for row in range(1, self.nRow):
                agent = self.get(row, self.colFake)
                if not agent.isFake():
                    agentNeig = self.get(row-1, self.colFake-1)
                    agent.setBorder(Dir.NORTH, agentNeig.getBorder(Dir.SOUTH))

    def fixSeg(self, width, length):
        w_unit = (self.imgW/self.nCol)/100
        l_unit = (self.imgH/self.nRow)/100
        w_side = round(width/2*w_unit)
        l_side = round(length/2*l_unit)
        for row in range(self.nRow):
            for col in range(self.nCol):
                agent = self.get(row, col)
                if not agent or agent.isFake():
                    continue
                agent.resetBorder()
                # set border
                if (row == (self.nRow-1)) & (length == 100):
                    self.setBorder(agent, Dir.SOUTH, self.imgH)
                else:
                    self.setBorder(agent, Dir.SOUTH, agent.y+l_side)
                if (row == 0) & (length == 100):
                    self.setBorder(agent, Dir.NORTH, 0)
                else:
                    self.setBorder(agent, Dir.NORTH, agent.y-l_side)
                if (col == (self.nCol-1)) & (width == 100):
                    self.setBorder(agent, Dir.EAST, self.imgW)
                else:
                    self.setBorder(agent, Dir.EAST, agent.x+w_side)
                if (col == 0) & (width == 100):
                    self.setBorder(agent, Dir.WEST, 0)
                else:
                    self.setBorder(agent, Dir.WEST, agent.x-w_side)

    def setCoordinate(self, agent, x, y):
        agent.setCoordinate(x, y)
        self.checkBorder(agent)

    def updateCoordinate(self, agent, value, axis):
        agent.updateCoordinate(value, axis)
        self.checkBorder(agent)

    def setBorder(self, agent, dir, value):
        agent.setBorder(dir, value)
        self.checkBorder(agent)
    
    def updateBorder(self, agent, dir, value):
        agent.updateBorder(dir, value)
        self.checkBorder(agent)

    def resetCoordinate(self):
        for row in range(self.nRow):
            for col in range(self.nCol):
                try:
                    self.get(row=row, col=col).resetCoordinate()
                except:
                    None

    def resetBorder(self):
        for row in range(self.nRow):
            for col in range(self.nCol):
                try:
                    self.get(row=row, col=col).resetBorder()
                except:
                    None

    def checkBorder(self, agent):
        if agent.border[Dir.NORTH.name] < 0:
            agent.border[Dir.NORTH.name] = 0
        if agent.border[Dir.WEST.name] < 0:
            agent.border[Dir.WEST.name] = 0
        if agent.border[Dir.SOUTH.name] >= self.imgH:
            agent.border[Dir.SOUTH.name] = self.imgH-1
        if agent.border[Dir.EAST.name] >= self.imgW:
            agent.border[Dir.EAST.name] = self.imgW-1

    def align(self, method, axis=0):
        '''
        '''
        if method == 0:
            # None
            for row in range(self.nRow):
                for col in range(self.nCol):
                    agent = self.get(row=row, col=col)
                    if not agent or agent.isFake():
                        continue
                    if axis == 0:
                        dist = agent.y_reset - agent.y
                    elif axis == 1:
                        dist = agent.x_reset - agent.x
                    self.updateCoordinate(agent, dist, axis=axis)
        elif method == 1:
            # Top/Left
            if axis == 0:
                for row in range(self.nRow):
                    val = 1e10
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        val_temp = agent.y
                        val = val_temp if val_temp < val else val
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = val - agent.y
                        self.updateCoordinate(agent, dist, axis=axis)
            elif axis == 1:
                for col in range(self.nCol):
                    val = 1e10
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        val_temp = agent.x
                        val = val_temp if val_temp < val else val
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = val - agent.x
                        self.updateCoordinate(agent, dist, axis=axis)
        elif method == 3:
            # Bottom/Right
            if axis == 0:
                for row in range(self.nRow):
                    val = -1
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        val_temp = agent.y
                        val = val_temp if val_temp > val else val
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = val - agent.y
                        self.updateCoordinate(agent, dist, axis=axis)
            elif axis == 1:
                for col in range(self.nCol):
                    val = -1
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        val_temp = agent.x
                        val = val_temp if val_temp > val else val
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = val - agent.x
                        self.updateCoordinate(agent, dist, axis=axis)
        elif method == 2:
            # Middle/Center
            if axis == 0:
                for row in range(self.nRow):
                    val = 0
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        val += agent.y
                    val = int(val/(self.nCol))
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = val - agent.y
                        self.updateCoordinate(agent, dist, axis=axis)
            elif axis == 1:
                for col in range(self.nCol):
                    val = 0
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        val += agent.x
                    val = int(val/(self.nRow))
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = val - agent.x
                        self.updateCoordinate(agent, dist, axis=axis)

    def pan(self, axis, target, value):
        if axis == 0:
            for col in range(self.nCol):
                agent = self.get(row=target, col=col)
                dist = value - agent.y
                self.updateCoordinate(agent, dist, axis=axis)
        elif axis == 1:
            for row in range(self.nRow):
                agent = self.get(row=row, col=target)
                dist = value - agent.x
                self.updateCoordinate(agent, dist, axis=axis)

    def distributed(self, axis, isEven):
        if isEven:
            if axis == 0: # row
                y_North = self.get(row=0, col=0).y
                y_South = self.get(row=self.nRow-1, col=0).y
                dist = y_South-y_North
                pos_new = np.arange(y_North, y_South, dist/(self.nRow-1))
                pos_new = np.append(pos_new, y_South)
                for row in range(self.nRow):
                    for col in range(self.nCol):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = pos_new[row] - agent.y
                        self.updateCoordinate(agent=agent, value=dist, axis=0)
            else:       # col
                x_West = self.get(row=0, col=0).x
                x_East = self.get(row=0, col=self.nCol-1).x
                dist = x_East-x_West
                pos_new = np.arange(x_West, x_East, dist/(self.nCol-1))
                pos_new = np.append(pos_new, x_East)
                for col in range(self.nCol):
                    for row in range(self.nRow):
                        agent = self.get(row=row, col=col)
                        if not agent or agent.isFake():
                            continue
                        dist = pos_new[col] - agent.x
                        self.updateCoordinate(agent=agent, value=dist, axis=1)
        else:
            for row in range(self.nRow):
                for col in range(self.nCol):
                    agent = self.get(row=row, col=col)
                    if not agent or agent.isFake():
                        continue
                    if axis == 0:
                        dist = agent.y_reset - agent.y
                    elif axis == 1:
                        dist = agent.x_reset - agent.x
                    self.updateCoordinate(agent=agent, value=dist, axis=axis)


class Agent():
    def __init__(self, name, row, col):
        '''
        '''
        self.name = name
        self.row, self.col = row, col
        self.y, self.x = 0, 0
        self.y_reset, self.x_reset = 0, 0
        self.pre_rg_W, self.pre_rg_H = range(0), range(0)
        self.border, self.border_reset = dict(), dict()
        for dir in list([Dir.NORTH, Dir.EAST, Dir.SOUTH, Dir.WEST]):
            self.border[dir.name] = 0
            self.border_reset[dir.name] = 0

    def getCoordinate(self):
        '''
        '''
        return self.x, self.y

    def getPreDim(self, isHeight=True):
        '''
        '''
        return self.pre_rg_H if isHeight else self.pre_rg_W

    def getBorder(self, dir):
        return self.border[dir.name]

    def getScoreArea(self, dir, img):
        '''
        Will ragne from 0 to 1
        '''
        isH = dir.value%2 # E->1, S->0
        rg = self.getPreDim(isHeight=isH)
        bd = self.getBorder(dir)
        return img[rg, bd].mean() if isH else img[bd, rg].mean()

    def getScoreGrid(self, dir):
        '''
        Will ragne from 0 to 1
        '''
        isWE = dir.value%2 # is W, E or N, S
        pt_center = self.x if isWE else self.y
        pt_cur = self.getBorder(dir)
        return abs(pt_cur-pt_center)

    def setCoordinate(self, x, y):
        '''
        '''
        self.x, self.y = int(x), int(y)
        self.x_reset, self.y_reset = int(x), int(y)
        self.setBorder(Dir.NORTH, y)
        self.setBorder(Dir.SOUTH, y)
        self.setBorder(Dir.WEST, x)
        self.setBorder(Dir.EAST, x)

    def setPreDim(self, rg):
        '''
        '''
        self.pre_rg_W = range(rg['WEST'], rg['EAST'])
        self.pre_rg_H = range(rg['NORTH'], rg['SOUTH'])
        self.x = int((rg['EAST']+rg['WEST'])/2)
        self.y = int((rg['NORTH']+rg['SOUTH'])/2)
        self.x_reset, self.y_reset = self.x, self.y
        for dir in list([Dir.NORTH, Dir.WEST, Dir.SOUTH, Dir.EAST]):
            self.border_reset[dir.name] = self.border[dir.name]

    def setBorder(self, dir, value):
        '''
        '''
        self.border[dir.name] = int(value)

    def updateBorder(self, dir, value):
        '''
        '''
        self.border[dir.name] += int(value)

    def updateCoordinate(self, value, axis=0):
        '''
        '''
        value = int(value)
        if axis==0:
            self.y += value
            self.border[Dir.NORTH.name] += value
            self.border[Dir.SOUTH.name] += value
        elif axis==1:
            self.x += value
            self.border[Dir.WEST.name] += value
            self.border[Dir.EAST.name] += value
    
    def getQRect(self):
        """
        GUI SPECIFIC
        """
        x = self.getBorder(Dir.WEST)
        y = self.getBorder(Dir.NORTH)
        w = self.getBorder(Dir.EAST) - x
        h = self.getBorder(Dir.SOUTH) - y
        return QRect(x, y, w, h)

    def resetCoordinate(self):
        self.x = self.x_reset
        self.y = self.y_reset
        self.resetBorder()

    def resetBorder(self):
        for dir in list([Dir.NORTH, Dir.WEST, Dir.SOUTH, Dir.EAST]):
            self.border[dir.name] = self.border_reset[dir.name]

    def resetBorderHard(self):
        self.setBorder(Dir.WEST, self.x)
        self.setBorder(Dir.EAST, self.x)
        self.setBorder(Dir.NORTH, self.y)
        self.setBorder(Dir.SOUTH, self.y)

    def isFake(self):
        return self.name=="_FAKE"
